"""
Module for VLT FORS (1 and 2)

.. include:: ../include/links.rst
"""
import glob
from pkg_resources import resource_filename

import numpy as np

from pypeit import msgs
from pypeit import telescopes
from pypeit.core import parse
from pypeit.core import framematch
from pypeit.spectrographs import spectrograph
from pypeit.images import detector_container


class VLTFORSSpectrograph(spectrograph.Spectrograph):
    """
    Child of Spectrograph to handle VLT/FORS specific code
    Parent for FORS1 and FORS2
    """
    ndet = 1  # Because each detector is written to a separate FITS file
    telescope = telescopes.VLTTelescopePar()

    @classmethod
    def default_pypeit_par(cls):
        """
        Return the default parameters to use for this instrument.
        
        Returns:
            :class:`~pypeit.par.pypeitpar.PypeItPar`: Parameters required by
            all of ``PypeIt`` methods.
        """
        par = super().default_pypeit_par()

        # Always correct for flexure, starting with default parameters
        par['flexure']['spec_method'] = 'boxcar'

        # Median overscan
        #   IF YOU CHANGE THIS, YOU WILL NEED TO DEAL WITH THE OVERSCAN GOING ALONG ROWS
        for key in par['calibrations'].keys():
            if 'frame' in key:
                par['calibrations'][key]['process']['overscan_method'] = 'median'

        # Adjustments to slit and tilts for NIR
        par['calibrations']['slitedges']['edge_thresh'] = 50.
        par['calibrations']['slitedges']['fit_order'] = 3
        par['calibrations']['slitedges']['max_shift_adj'] = 0.5

        # Tilt parameters
        par['calibrations']['tilts']['tracethresh'] = 25.0
        par['calibrations']['tilts']['spat_order'] = 3
        par['calibrations']['tilts']['spec_order'] = 4

        # 1D wavelength solution
        par['calibrations']['wavelengths']['lamps'] = ['HeI', 'ArI']  # Grating dependent
        par['calibrations']['wavelengths']['rms_threshold'] = 0.25
        par['calibrations']['wavelengths']['sigdetect'] = 10.0
        par['calibrations']['wavelengths']['fwhm'] = 4.0  # Good for 2x binning
        par['calibrations']['wavelengths']['n_final'] = 4

        # Flats
        par['calibrations']['flatfield']['tweak_slits_thresh'] = 0.90
        par['calibrations']['flatfield']['tweak_slits_maxfrac'] = 0.10

        return par

    def init_meta(self):
        """
        Define how metadata are derived from the spectrograph files.

        That is, this associates the ``PypeIt``-specific metadata keywords
        with the instrument-specific header cards using :attr:`meta`.
        """
        self.meta = {}
        # Required (core)
        self.meta['ra'] = dict(ext=0, card='RA', required_ftypes=['science', 'standard'])  # Need to convert to : separated
        self.meta['dec'] = dict(ext=0, card='DEC', required_ftypes=['science', 'standard'])
        self.meta['target'] = dict(ext=0, card='OBJECT')
        self.meta['binning'] = dict(card=None, compound=True)

        self.meta['mjd'] = dict(ext=0, card='MJD-OBS')
        self.meta['exptime'] = dict(ext=0, card='EXPTIME')
        self.meta['airmass'] = dict(ext=0, card='HIERARCH ESO TEL AIRM START', required_ftypes=['science', 'standard'])
        #
        self.meta['decker'] = dict(card=None, compound=True, required_ftypes=['science', 'standard'])
        # Extras for config and frametyping
        self.meta['dispname'] = dict(ext=0, card='HIERARCH ESO INS GRIS1 NAME', required_ftypes=['science', 'standard'])
        self.meta['dispangle'] = dict(ext=0, card='HIERARCH ESO INS GRIS1 WLEN', rtol=2.0, required_ftypes=['science', 'standard'])
        self.meta['idname'] = dict(ext=0, card='HIERARCH ESO DPR CATG')
        self.meta['detector'] = dict(ext=0, card='EXTNAME')
        self.meta['instrument'] = dict(ext=0, card='INSTRUME')

    def compound_meta(self, headarr, meta_key):
        """
        Methods to generate metadata requiring interpretation of the header
        data, instead of simply reading the value of a header card.

        Args:
            headarr (:obj:`list`):
                List of `astropy.io.fits.Header`_ objects.
            meta_key (:obj:`str`):
                Metadata keyword to construct.

        Returns:
            object: Metadata value read from the header(s).
        """
        if meta_key == 'binning':
            binspatial = headarr[0]['HIERARCH ESO DET WIN1 BINX']
            binspec = headarr[0]['HIERARCH ESO DET WIN1 BINY']
            binning = parse.binning2string(binspec, binspatial)
            return binning
        elif meta_key == 'decker':
            try:  # Science
                decker = headarr[0]['HIERARCH ESO INS SLIT NAME']
            except KeyError:  # Standard!
                try:
                    decker = headarr[0]['HIERARCH ESO SEQ SPEC TARG']
                except KeyError:
                    return None
            return decker
        else:
            msgs.error("Not ready for this compound meta")

    def configuration_keys(self):
        """
        Return the metadata keys that define a unique instrument
        configuration.

        This list is used by :class:`~pypeit.metadata.PypeItMetaData` to
        identify the unique configurations among the list of frames read
        for a given reduction.

        Returns:
            :obj:`list`: List of keywords of data pulled from file headers
            and used to constuct the :class:`~pypeit.metadata.PypeItMetaData`
            object.
        """
        return []

    def check_frame_type(self, ftype, fitstbl, exprng=None):
        """
        Check for frames of the provided type.

        Args:
            ftype (:obj:`str`):
                Type of frame to check. Must be a valid frame type; see
                frame-type :ref:`frame_type_defs`.
            fitstbl (`astropy.table.Table`_):
                The table with the metadata for one or more frames to check.
            exprng (:obj:`list`, optional):
                Range in the allowed exposure time for a frame of type
                ``ftype``. See
                :func:`pypeit.core.framematch.check_frame_exptime`.

        Returns:
            `numpy.ndarray`_: Boolean array with the flags selecting the
            exposures in ``fitstbl`` that are ``ftype`` type frames.
        """
        good_exp = framematch.check_frame_exptime(fitstbl['exptime'], exprng)
        # TODO: Allow for 'sky' frame type, for now include sky in
        # 'science' category
        if ftype == 'science':
            return good_exp & ((fitstbl['idname'] == 'SCIENCE')
                                | (fitstbl['target'] == 'STD,TELLURIC')
                                | (fitstbl['target'] == 'STD,SKY'))
        if ftype == 'standard':
            return good_exp & ((fitstbl['target'] == 'STD,FLUX')
                               | (fitstbl['target'] == 'STD'))
        if ftype == 'bias':
            return good_exp & (fitstbl['target'] == 'BIAS')
        if ftype == 'dark':
            return good_exp & (fitstbl['target'] == 'DARK')
        if ftype in ['pixelflat', 'trace', 'illumflat']:
            # Flats and trace frames are typed together
            return good_exp & ((fitstbl['target'] == 'LAMP,DFLAT')
                               | (fitstbl['target'] == 'LAMP,QFLAT')
                               | (fitstbl['target'] == 'FLAT,LAMP')
                               | (fitstbl['target'] == 'LAMP,FLAT'))
        if ftype == 'pinhole':
            # Don't type pinhole
            return np.zeros(len(fitstbl), dtype=bool)
        if ftype in ['arc', 'tilt']:
            return good_exp & ((fitstbl['target'] == 'LAMP,WAVE')
                               | (fitstbl['target'] == 'WAVE,LAMP'))

        msgs.warn('Cannot determine if frames are of type {0}.'.format(ftype))
        return np.zeros(len(fitstbl), dtype=bool)


class VLTFORS2Spectrograph(VLTFORSSpectrograph):
    """
    Child to handle VLT/FORS2 specific code
    """

    name = 'vlt_fors2'
    camera = 'FORS2'
    header_name = 'FORS2'
    supported = True
    comment = '300I, 300V gratings'

    def get_detector_par(self, det, hdu=None):
        """
        Return metadata for the selected detector.

        Args:
            det (:obj:`int`):
                1-indexed detector number.  ESO writes each of the two detectors
                to separate files.  When ``hdu`` is provided, this is ignored
                and instead the chip is determined by the header parameter
                "EXTNAME".  If ``hdu`` is None (for automatically generated
                documentation only), this can be used to set the chip (1 or 2)
                that is returned.
            hdu (`astropy.io.fits.HDUList`_, optional):
                The open fits file with the raw image of interest.  If not
                provided, frame-dependent parameters are set to a default.

        Returns:
            :class:`~pypeit.images.detector_container.DetectorContainer`:
            Object with the detector metadata.
        """
        if hdu is None:
            binning = '1,1'
            chip = 'CHIP1' if det == 1 else 'CHIP2'
        else:
            # Binning
            # TODO: Could this be detector dependent??
            binning = self.get_meta_value(self.get_headarr(hdu), 'binning')
            chip = self.get_meta_value(self.get_headarr(hdu), 'detector')

        # Detector 1 (Thor)  -- http://www.eso.org/sci/php/optdet/instruments/fors2/index.html
        detector_dict1 = dict(
            binning         = binning,
            det             = 1,
            dataext         = 0,
            specaxis        = 1,
            specflip        = False,
            spatflip        = False,
            platescale      = 0.126,  # average between order 11 & 30, see manual
            darkcurr        = 0.0,
            saturation      = 2.0e5,  # I think saturation may never be a problem here since there are many DITs
            nonlinear       = 0.80,
            mincounts       = -1e10,
            numamplifiers   = 1,
            gain            = np.atleast_1d(0.70),
            ronoise         = np.atleast_1d(2.9), # High gain
            datasec         = np.atleast_1d('[11:2059,:]'),  # For 1x binning, I think
            oscansec        = np.atleast_1d('[2062:,:]'),
            #suffix          = '_Thor',
        )
        # Detector 2 (Belenos)
        detector_dict2 = dict(
            binning         = binning,
            det             = 1,  # ESO writes these to separate FITS images!!
            dataext         = 0,
            specaxis        = 1,
            specflip        = False,
            spatflip        = False,
            platescale      = 0.126,  # average between order 11 & 30, see manual
            darkcurr        = 0.0,
            saturation      = 2.0e5,  # I think saturation may never be a problem here since there are many DITs
            nonlinear       = 0.80,
            mincounts       = -1e10,
            numamplifiers   = 1,
            gain            = np.atleast_1d(0.70),
            ronoise         = np.atleast_1d(3.0),  # High gain
            datasec         = np.atleast_1d('[20:,0:2048]'),
            oscansec        = np.atleast_1d('[4:20,4:2044]'),
            #suffix          = '_Belenos'
        )

        # Finish
        if chip == 'CHIP1':
            return detector_container.DetectorContainer(**detector_dict1)
        elif chip == 'CHIP2':
            return detector_container.DetectorContainer(**detector_dict2)
        else:
            msgs.error(f'Unknown chip: {chip}!')

    def config_specific_par(self, scifile, inp_par=None):
        """
        Modify the ``PypeIt`` parameters to hard-wired values used for
        specific instrument configurations.

        Args:
            scifile (:obj:`str`):
                File to use when determining the configuration and how
                to adjust the input parameters.
            inp_par (:class:`~pypeit.par.parset.ParSet`, optional):
                Parameter set used for the full run of PypeIt.  If None,
                use :func:`default_pypeit_par`.

        Returns:
            :class:`~pypeit.par.parset.ParSet`: The PypeIt parameter set
            adjusted for configuration specific parameter values.
        """
        # Start with instrument wide
        par = super().config_specific_par(scifile, inp_par=inp_par)
        # TODO: Should we allow the user to override these?

        #detector = self.get_meta_value(scifile, 'detector')
        #self.set_detector(detector)
        # Wavelengths
        #par['calibrations']['wavelengths']['nonlinear_counts'] = self.detector[0]['nonlinear'] * self.detector[0]['saturation']

        if self.get_meta_value(scifile, 'dispname') == 'GRIS_300I':
            par['calibrations']['wavelengths']['reid_arxiv'] = 'vlt_fors2_300I.fits'
            par['calibrations']['wavelengths']['method'] = 'full_template'
        elif self.get_meta_value(scifile, 'dispname') == 'GRIS_300V':
            par['calibrations']['wavelengths']['reid_arxiv'] = 'vlt_fors2_300V.fits'
            par['calibrations']['wavelengths']['method'] = 'full_template'

        if 'lSlit' in self.get_meta_value(scifile, 'decker'):
            par['calibrations']['slitedges']['sync_predict'] = 'nearest'

        return par

    def configuration_keys(self):
        """
        Return the metadata keys that define a unique instrument
        configuration.

        This list is used by :class:`~pypeit.metadata.PypeItMetaData` to
        identify the unique configurations among the list of frames read
        for a given reduction.

        Returns:
            :obj:`list`: List of keywords of data pulled from file headers
            and used to constuct the :class:`~pypeit.metadata.PypeItMetaData`
            object.
        """
        return ['dispname', 'dispangle', 'decker', 'detector']





""" Module for finding patterns in arc line spectra

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst

"""
import numpy as np
import inspect

from astropy.io import fits

from pypeit.core.wavecal import autoid
from pypeit.core.wavecal import defs
from pypeit.core import fitting
from pypeit import msgs

from pypeit import datamodel

from IPython import embed


class WaveFit(datamodel.DataContainer):
    """
    DataContainer for the output from BuildWaveCalib

    All of the items in the datamodel are required for instantiation,
      although they can be None (but shouldn't be)

    """
    version = '1.0.0'

    datamodel = {'spat_id': dict(otype=(int,np.integer), descr='Spatial position of slit/order for this fit. Required for I/O'),
                 'pypeitfit': dict(otype=fitting.PypeItFit,
                                   descr='Fit to 1D wavelength solutions'),
                 'pixel_fit': dict(otype=np.ndarray, atype=np.floating,
                                   descr='Pixel values of arc lines'),
                 'wave_fit': dict(otype=np.ndarray, atype=np.floating,
                                  descr='Wavelength IDs assigned'),
                 'xnorm': dict(otype=float, descr='Normalization for fit'),
                 'ion_bits': dict(otype=np.ndarray, atype=np.integer,
                                  descr='Ion bit values for the Ion names'),
                 'cen_wave': dict(otype=float, descr='Central wavelength'),
                 'cen_disp': dict(otype=float, descr='Approximate wavelength dispersion'),
                 'spec': dict(otype=np.ndarray, atype=np.floating, descr='Arc spectrum'),
                 'wave_soln': dict(otype=np.ndarray, atype=np.floating,
                                   descr='Evaluated wavelengths at pixel_fit'),
                 'sigrej': dict(otype=float, descr='Final sigma rejection applied'),
                 'shift': dict(otype=float, descr='Shift applied'),
                 'tcent': dict(otype=np.ndarray, atype=np.floating,
                               descr='Pixel centroids of all arc lines found'),
                 'rms': dict(otype=float, descr='RMS of the solution')}

    bitmask = defs.LinesBitMask()

    @staticmethod
    def hduext_prefix_from_spatid(spat_id):
        """ Naming for HDU extensions"""
        return 'SPAT_ID-{}_'.format(spat_id)

    def __init__(self, spat_id, pypeitfit=None, pixel_fit=None, wave_fit=None, ion_bits=None,
                 cen_wave=None, cen_disp=None, spec=None, wave_soln=None,
                 sigrej=None, shift=None, tcent=None, rms=None, xnorm=None):
        # Parse
        args, _, _, values = inspect.getargvalues(inspect.currentframe())
        d = dict([(k,values[k]) for k in args[1:]])
        # Setup the DataContainer
        datamodel.DataContainer.__init__(self, d=d)

    def _bundle(self, **kwargs):
        """
        Over-ride DataContainer._bundle() to deal with PYPEITFIT

        Args:
            kwargs:
                Passed to DataContainer._bundle()

        Returns:
            list:

        """
        # Extension prefix (for being unique with slits)
        hdue_pref = self.hduext_prefix_from_spatid(self.spat_id)
        # Without PypeItFit
        _d = super(WaveFit, self)._bundle(
            ext=hdue_pref+'WAVEFIT', **kwargs)
        # Deal with PypeItFit
        if _d[0][hdue_pref+'WAVEFIT']['pypeitfit'] is not None:
            _d.append({hdue_pref+'PYPEITFIT': _d[0][hdue_pref + 'WAVEFIT'].pop('pypeitfit')})
        # Return
        return _d

    def to_hdu(self, hdr=None, add_primary=False, primary_hdr=None, limit_hdus=None):
        """ Over-ride for force_to_bintbl

        See :class:`pypeit.datamodel.DataContainer.to_hdu` for Arguments

        Returns:
            :obj:`list`, `astropy.io.fits.HDUList`_: A list of HDUs,
            where the type depends on the value of ``add_primary``.
        """
        return super(WaveFit, self).to_hdu(hdr=hdr, add_primary=add_primary, primary_hdr=primary_hdr,
                                           limit_hdus=limit_hdus, force_to_bintbl=True)

    @classmethod
    def from_hdu(cls, hdu, chk_version=True):
        """
        Parse the data from the provided HDU.

        See :func:`pypeit.datamodel.DataContainer._parse` for the
        argument descriptions.
        """
        # Set hdu_prefix
        if isinstance(hdu, fits.HDUList):
            hdu_prefix = cls.hduext_prefix_from_spatid(hdu[1].header['SPAT_ID'])
        else:
            hdu_prefix = cls.hduext_prefix_from_spatid(hdu.header['SPAT_ID'])
        # Run the default parser to get the data
        return super(WaveFit, cls).from_hdu(hdu, hdu_prefix=hdu_prefix)

    @property
    def ions(self):
        """
        Returns an array of ion labels

        Returns:
            `numpy.ndarray`_:  Array of the ion label for each line as recorded in ion_bits

        """
        ionlist = []
        for ionbit in self.ion_bits:
            ionlist += self.bitmask.flagged_bits(ionbit)
        # Return
        return np.asarray(ionlist)


def fit_slit(spec, patt_dict, tcent, line_lists, vel_tol = 1.0, outroot=None, slittxt="Slit", thar=False,match_toler=3.0,
             func='legendre', n_first=2,sigrej_first=2.0,n_final=4,sigrej_final=3.0,verbose=False):

    """ Perform a fit to the wavelength solution. Wrapper for iterative fitting code.

    Parameters
    ----------
    spec : ndarray
      arc spectrum
    patt_dict : dict
      dictionary of patterns
    tcent: ndarray
      List of the detections in this slit to be fit using the patt_dict
    line_lists: astropy Table
      Table containing the line list
    Optional Parameters
    -------------------
    vel_tol: float, default = 1.0
      Tolerance in km/s for matching lines in the IDs to lines in the NIST database. The default is 1.0 km/s
    outroot: str
      Path for QA file.
    slittxt : str
      Label used for QA
    thar: bool, default = False
      True if this is a ThAr fit
    match_toler: float, default = 3.0
      Matching tolerance when searching for new lines. This is the difference in pixels between the wavlength assigned to
      an arc line by an iteration of the wavelength solution to the wavelength in the line list.
    func: str, default = 'legendre'
      Name of function used for the wavelength solution
    n_first: int, default = 2
      Order of first guess to the wavelength solution.
    sigrej_first: float, default = 2.0
      Number of sigma for rejection for the first guess to the wavelength solution.
    n_final: int, default = 4
      Order of the final wavelength solution fit
    sigrej_final: float, default = 3.0
      Number of sigma for rejection for the final fit to the wavelength solution.
    verbose : bool
      If True, print out more information.
    plot_fil:
      Filename for plotting some QA?

    Returns
    -------
    final_fit : dict
      A dictionary containing all of the information about the fit
    """

    # Check that patt_dict and tcent refer to each other
    if patt_dict['mask'].shape != tcent.shape:
        msgs.error('patt_dict and tcent do not refer to each other. Something is very wrong')

    # Perform final fit to the line IDs
    if thar:
        NIST_lines = (line_lists['NIST'] > 0) & (np.char.find(line_lists['Source'].data, 'MURPHY') >= 0)
    else:
        NIST_lines = line_lists['NIST'] > 0
    ifit = np.where(patt_dict['mask'])[0]

    if outroot is not None:
        plot_fil = outroot + slittxt + '_fit.pdf'
    else:
        plot_fil = None

    # TODO Profx maybe you can add a comment on what this is doing. Why do we have use_unknowns=True only to purge them later??
    # Purge UNKNOWNS from ifit
    imsk = np.ones(len(ifit), dtype=np.bool)
    for kk, idwv in enumerate(np.array(patt_dict['IDs'])[ifit]):
        if (np.min(np.abs(line_lists['wave'][NIST_lines] - idwv)))/idwv*3.0e5 > vel_tol:
            imsk[kk] = False
    ifit = ifit[imsk]
    # Fit
    final_fit = iterative_fitting(spec, tcent, ifit, np.array(patt_dict['IDs'])[ifit], line_lists[NIST_lines],
                                  patt_dict['bdisp'],match_toler=match_toler, func=func, n_first=n_first,
                                  sigrej_first=sigrej_first,n_final=n_final, sigrej_final=sigrej_final,
                                  plot_fil=plot_fil, verbose=verbose)
    if plot_fil is not None and final_fit is not None:
        print("Wrote: {:s}".format(plot_fil))

    # Return
    return final_fit


def iterative_fitting(spec, tcent, ifit, IDs, llist, disp,
                      match_toler = 2.0, func = 'legendre', n_first=2, sigrej_first=2.0,
                      n_final=4, sigrej_final=3.0, input_only=False,
                      weights=None, plot_fil=None, verbose=False):

    """ Routine for iteratively fitting wavelength solutions.

    Parameters
    ----------
    spec : ndarray, shape = (nspec,)
      arcline spectrum
    tcent : ndarray
      Centroids in pixels of lines identified in spec
    ifit : ndarray
      Indices of the lines that will be fit
    IDs: ndarray
      wavelength IDs of the lines that will be fit (I think?)
    llist: dict
      Linelist dictionary
    disp: float
      dispersion

    Optional Parameters
    -------------------
    match_toler: float, default = 3.0
      Matching tolerance when searching for new lines. This is the difference in pixels between the wavlength assigned to
      an arc line by an iteration of the wavelength solution to the wavelength in the line list.
    func: str, default = 'legendre'
      Name of function used for the wavelength solution
    n_first: int, default = 2
      Order of first guess to the wavelength solution.
    sigrej_first: float, default = 2.0
      Number of sigma for rejection for the first guess to the wavelength solution.
    n_final: int, default = 4
      Order of the final wavelength solution fit
    sigrej_final: float, default = 3.0
      Number of sigma for rejection for the final fit to the wavelength solution.
    input_only: bool
      If True, the routine will only perform a robust polyfit to the input IDs.
      If False, the routine will fit the input IDs, and then include additional
      lines in the linelist that are a satisfactory fit.
    weights: ndarray
      Weights to be used?
    verbose : bool
      If True, print out more information.
    plot_fil:
      Filename for plotting some QA?

    Returns
    -------
    final_fit: :class:`pypeit.core.wavecal.wv_fitting.WaveFit`
    """

    #TODO JFH add error checking here to ensure that IDs and ifit have the same size!

    if weights is None:
        weights = np.ones(tcent.size)

    nspec = spec.size
    xnspecmin1 = float(nspec-1)
    # Setup for fitting
    sv_ifit = list(ifit)  # Keep the originals
    all_ids = -999.*np.ones(len(tcent))
    all_idsion = np.array(['UNKNWN']*len(tcent))
    all_ids[ifit] = IDs

    # Fit
    n_order = n_first
    flg_continue = True
    flg_penultimate = False
    fmin, fmax = 0.0, 1.0
    # Note the number of parameters is actually n_order and not n_order+1
    while flg_continue:
        if flg_penultimate:
            flg_continue = False
        # Fit with rejection
        xfit, yfit, wfit = tcent[ifit], all_ids[ifit], weights[ifit]
        maxiter = xfit.size - n_order - 2
        #
        if xfit.size == 0:
            msgs.warn("All points rejected !!")
            return None
        # Fit
        pypeitFit = fitting.robust_fit(xfit/xnspecmin1, yfit, n_order, function=func, maxiter=maxiter,
                                       lower=sigrej_first, upper=sigrej_first, maxrej=1, sticky=True,
                                       minx=fmin, maxx=fmax, weights=wfit)
        # Junk fit?
        if pypeitFit is None:
            msgs.warn("Bad fit!!")
            return None

        rms_ang = pypeitFit.calc_fit_rms(apply_mask=True)
        rms_pix = rms_ang/disp
        if verbose:
            msgs.info('n_order = {:d}'.format(n_order) + ': RMS = {:g}'.format(rms_pix))

        # Reject but keep originals (until final fit)
        ifit = list(ifit[pypeitFit.gpm == 1]) + sv_ifit
        if not input_only:
            # Find new points from the linelist (should we allow removal of the originals?)
            twave = pypeitFit.eval(tcent/xnspecmin1)#, func, minx=fmin, maxx=fmax)
            for ss, iwave in enumerate(twave):
                mn = np.min(np.abs(iwave-llist['wave']))
                if mn/disp < match_toler:
                    imn = np.argmin(np.abs(iwave-llist['wave']))
                    #if verbose:
                    #    print('Adding {:g} at {:g}'.format(llist['wave'][imn],tcent[ss]))
                    # Update and append
                    all_ids[ss] = llist['wave'][imn]
                    all_idsion[ss] = llist['ion'][imn]
                    ifit.append(ss)
        # Keep unique ones
        ifit = np.unique(np.array(ifit, dtype=int))
        # Increment order?
        if n_order < n_final:
            n_order += 1
        else:
            flg_penultimate = True

    # Final fit (originals can now be rejected)
    xfit, yfit, wfit = tcent[ifit], all_ids[ifit], weights[ifit]
    pypeitFit = fitting.robust_fit(xfit/xnspecmin1, yfit, n_order, function=func,
                                   lower=sigrej_final, upper=sigrej_final, maxrej=1, sticky=True,
                                   minx=fmin, maxx=fmax, weights=wfit)#, debug=True)
    irej = np.where(np.logical_not(pypeitFit.bool_gpm))[0]
    if len(irej) > 0:
        xrej = xfit[irej]
        yrej = yfit[irej]
        if verbose:
            for kk, imask in enumerate(irej):
                wave = pypeitFit.eval(xrej[kk]/xnspecmin1)#, func, minx=fmin, maxx=fmax)
                msgs.info('Rejecting arc line {:g}; {:g}'.format(yfit[imask], wave))
    else:
        xrej = []
        yrej = []

    ions = all_idsion[ifit]
    # Final RMS
    rms_ang = pypeitFit.calc_fit_rms(apply_mask=True)
    rms_pix = rms_ang/disp

    # Pack up fit
    spec_vec = np.arange(nspec)
    wave_soln = pypeitFit.eval(spec_vec/xnspecmin1)
    cen_wave = pypeitFit.eval(float(nspec)/2/xnspecmin1)
    cen_wave_min1 = pypeitFit.eval((float(nspec)/2 - 1.0)/xnspecmin1)
    cen_disp = cen_wave - cen_wave_min1

    # Ions bit
    ion_bits = np.zeros(len(ions), dtype=WaveFit.bitmask.minimum_dtype())
    for kk,ion in enumerate(ions):
        ion_bits[kk] = WaveFit.bitmask.turn_on(ion_bits[kk], ion.replace(' ', ''))

    # DataContainer time
    # spat_id is set to an arbitrary -1 here and is updated in wavecalib.py
    final_fit = WaveFit(-1, pypeitfit=pypeitFit, pixel_fit=xfit, wave_fit=yfit,
                        ion_bits=ion_bits, xnorm=xnspecmin1,
                        cen_wave=cen_wave, cen_disp=cen_disp,
                        spec=spec, wave_soln = wave_soln, sigrej=sigrej_final,
                        shift=0., tcent=tcent, rms=rms_pix)

    # QA
    if plot_fil is not None:
        autoid.arc_fit_qa(final_fit, plot_fil)
    # Return
    return final_fit



from pypeit.utils import all_subclasses
from pypeit.spectrographs import spectrograph

# The import of all the spectrograph modules here is what enables the dynamic
# compiling of all the available spectrographs below
from pypeit.spectrographs import gtc_osiris
from pypeit.spectrographs import bok_bc
from pypeit.spectrographs import gemini_flamingos
from pypeit.spectrographs import gemini_gmos
from pypeit.spectrographs import gemini_gnirs
from pypeit.spectrographs import keck_deimos
from pypeit.spectrographs import keck_hires
from pypeit.spectrographs import keck_kcwi
from pypeit.spectrographs import keck_lris
from pypeit.spectrographs import keck_mosfire
from pypeit.spectrographs import keck_nires
from pypeit.spectrographs import keck_nirspec
from pypeit.spectrographs import lbt_luci
from pypeit.spectrographs import lbt_mods
from pypeit.spectrographs import ldt_deveny
from pypeit.spectrographs import magellan_fire
from pypeit.spectrographs import magellan_mage
from pypeit.spectrographs import mdm_osmos
from pypeit.spectrographs import mmt_binospec
from pypeit.spectrographs import mmt_bluechannel
from pypeit.spectrographs import mmt_mmirs
from pypeit.spectrographs import not_alfosc
from pypeit.spectrographs import p200_dbsp
from pypeit.spectrographs import p200_tspec
from pypeit.spectrographs import shane_kast
from pypeit.spectrographs import soar_goodman
from pypeit.spectrographs import tng_dolores
from pypeit.spectrographs import vlt_fors
from pypeit.spectrographs import vlt_xshooter
from pypeit.spectrographs import vlt_sinfoni
from pypeit.spectrographs import wht_isis
from pypeit.spectrographs import ntt_efosc2


# Build the list of names for the available spectrographs

def spectrograph_classes():
    import numpy as np
    # Recursively collect all subclasses
    spec_c = np.array(list(all_subclasses(spectrograph.Spectrograph)))
    # Select spectrograph classes with a defined name; spectrographs without a
    # name are either undefined or a base class.
    spec_c = spec_c[[c.name is not None for c in spec_c]]
    # Construct a dictionary with the spectrograph name and class
    srt = np.argsort(np.array([c.name for c in spec_c]))
    return dict([ (c.name,c) for c in spec_c[srt]])

available_spectrographs = list(spectrograph_classes().keys())



"""
Implements the flat-field class.

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst

"""
import copy
import inspect
import numpy as np

from scipy import interpolate

from matplotlib import pyplot as plt

from IPython import embed

from pypeit import msgs
from pypeit import utils
from pypeit import bspline

from pypeit import datamodel
from pypeit import masterframe
from pypeit.display import display
from pypeit.core import flat
from pypeit.core import tracewave
from pypeit.core import basis
from pypeit.core import fitting
from pypeit.core import coadd
from pypeit import slittrace


class FlatImages(datamodel.DataContainer):
    """
    Simple DataContainer for the output from Flatfield

    All of the items in the datamodel are required for instantiation,
      although they can be None (but shouldn't be)

    """
    minimum_version = '1.1.0'
    version = '1.1.0'

    # I/O
    output_to_disk = None  # This writes all items that are not None
    hdu_prefix = None

    # Master fun
    master_type = 'Flat'
    master_file_format = 'fits'

    datamodel = {'pixelflat_raw': dict(otype=np.ndarray, atype=np.floating,
                                       descr='Processed, combined pixel flats'),
                 'pixelflat_norm': dict(otype=np.ndarray, atype=np.floating,
                                        descr='Normalized pixel flat'),
                 'pixelflat_model': dict(otype=np.ndarray, atype=np.floating, descr='Model flat'),
                 'pixelflat_spat_bsplines': dict(otype=np.ndarray, atype=bspline.bspline,
                                                 descr='B-spline models for pixel flat'),
                 'pixelflat_bpm': dict(otype=np.ndarray, atype=np.integer,
                                       descr='Mirrors SlitTraceSet mask for Flat-specific flags'),
                 'pixelflat_spec_illum': dict(otype=np.ndarray, atype=np.floating,
                                              descr='Relative spectral illumination'),
                 'illumflat_raw': dict(otype=np.ndarray, atype=np.floating,
                                       descr='Processed, combined illum flats'),
                 'illumflat_spat_bsplines': dict(otype=np.ndarray, atype=bspline.bspline,
                                                 descr='B-spline models for illum flat'),
                 'illumflat_bpm': dict(otype=np.ndarray, atype=np.integer,
                                       descr='Mirrors SlitTraceSet mask for Flat-specific flags'),
                 'PYP_SPEC': dict(otype=str, descr='PypeIt spectrograph name'),
                 'spat_id': dict(otype=np.ndarray, atype=np.integer, descr='Slit spat_id')}

    def __init__(self, pixelflat_raw=None, pixelflat_norm=None, pixelflat_bpm=None,
                 pixelflat_model=None, pixelflat_spat_bsplines=None, pixelflat_spec_illum=None,
                 illumflat_raw=None, illumflat_spat_bsplines=None, illumflat_bpm=None,
                 PYP_SPEC=None, spat_id=None):
        # Parse
        args, _, _, values = inspect.getargvalues(inspect.currentframe())
        d = dict([(k,values[k]) for k in args[1:]])
        # Setup the DataContainer
        datamodel.DataContainer.__init__(self, d=d)

    def _init_internals(self):
        self.filename = None
        # Master stuff
        self.master_key = None
        self.master_dir = None

    def _validate(self):
        #
        if self.pixelflat_spat_bsplines is not None and len(self.pixelflat_spat_bsplines) > 0:
            if len(self.spat_id) != len(self.pixelflat_spat_bsplines):
                msgs.error("Pixelflat Bsplines are out of sync with the slit IDs")
        if self.illumflat_spat_bsplines is not None and len(self.illumflat_spat_bsplines) > 0:
            if len(self.spat_id) != len(self.illumflat_spat_bsplines):
                msgs.error("Illumflat Bsplines are out of sync with the slit IDs")

    def is_synced(self, slits):
        """
        Confirm the slits in WaveTilts are aligned to that in SlitTraceSet

        Barfs if not

        Args:
            slits (:class:`pypeit.slittrace.SlitTraceSet`):

        """
        if not np.array_equal(self.spat_id, slits.spat_id):
            msgs.error("Your flat solutions are out of sync with your slits.  Remove Masters and start from scratch")

    def _bundle(self):
        """
        Over-write default _bundle() method to write one
        HDU per image.  Any extras are in the HDU header of
        the primary image.

        Returns:
            :obj:`list`: A list of dictionaries, each list element is
            written to its own fits extension. See the description
            above.
        """
        d = []
        for key in self.keys():
            # Skip None
            if self[key] is None:
                continue
            if self.datamodel[key]['otype'] == np.ndarray and 'bsplines' not in key:
                d += [{key: self[key]}]
            elif 'bsplines' in key:
                flattype = 'pixelflat' if 'pixelflat' in key else 'illumflat'
                d += [{'{0}_spat_id-{1}_bspline'.format(flattype, self.spat_id[ss]): self[key][ss]}
                            for ss in range(len(self[key]))]
            else:
                if len(d) > 0:
                    d[0][key] = self[key]
                else:
                    d += [{key: self[key]}]
        # Return
        return d

    # TODO: Although I don't like doing it, kwargs is here to catch the
    # extraneous keywords that can be passed to _parse from the base class but
    # won't be used.
    @classmethod
    def _parse(cls, hdu, ext=None, transpose_table_arrays=False, hdu_prefix=None, **kwargs):

        # Grab everything but the bsplines. The bsplines are not parsed
        # because the tailored extension names do not match any of the
        # datamodel keys.
        d, version_passed, type_passed, parsed_hdus = super(FlatImages, cls)._parse(hdu)

        # Find bsplines, if they exist
        nspat = len(d['spat_id'])
        hdunames = [h.name for h in hdu]
        for flattype in ['pixelflat', 'illumflat']:
            ext = ['{0}_SPAT_ID-{1}_BSPLINE'.format(flattype.upper(), d['spat_id'][i])
                   for i in range(nspat)]
            indx = np.isin(ext, hdunames)
            if np.any(indx) and not np.all(indx):
                msgs.error('Expected {0} {1} bspline extensions, but only found {1}.'.format(
                           nspat, flattype, np.sum(indx)))
            if np.all(indx):
                key = '{0}_spat_bsplines'.format(flattype)
                try:
                    d[key] = np.array([bspline.bspline.from_hdu(hdu[k]) for k in ext])
                except Exception as e:
                    msgs.warn('Error in bspline extension read:\n {0}: {1}'.format(
                                e.__class__.__name__, str(e)))
                    # Assume this is because the type failed
                    type_passed = False
                else:
                    version_passed &= np.all([d[key][i].version == bspline.bspline.version 
                                              for i in range(nspat)])
                    parsed_hdus += ext

        return d, version_passed, type_passed, parsed_hdus

    def shape(self):
        if self.pixelflat_raw is not None:
            return self.pixelflat_raw.shape
        elif self.illumflat_raw is not None:
            return self.illumflat_raw.shape
        else:
            msgs.error("Shape of FlatImages could not be determined")

    def get_procflat(self, frametype='pixel'):
        if frametype == 'illum':
            return self.illumflat_raw
        else:
            return self.pixelflat_raw

    def get_bpmflats(self, frametype='pixel'):
        # Check if both BPMs are none
        if self.pixelflat_bpm is None and self.illumflat_bpm is None:
            msgs.warn("FlatImages contains no BPM - trying to generate one")
            return np.zeros(self.shape(), dtype=np.int)
        # Now return the requested case, checking for None
        if frametype == 'illum':
            if self.illumflat_bpm is not None:
                return self.illumflat_bpm
            else:
                msgs.warn("illumflat has no BPM - using the pixelflat BPM")
                return self.pixelflat_bpm
        else:
            if self.pixelflat_bpm is not None:
                return self.pixelflat_bpm
            else:
                msgs.warn("pixelflat has no BPM - using the illumflat BPM")
                return self.illumflat_bpm

    def get_spat_bsplines(self, frametype='illum'):
        """
        Grab a list of bspline fits

        Args:
            frametype (str):

        Returns:
            list:

        """
        # Check if both spat bsplines are none
        if self.pixelflat_spat_bsplines is None and self.illumflat_spat_bsplines is None:
            msgs.error("FlatImages contains no spatial bspline fit")
        # Now return the requested case, checking for None
        if frametype == 'illum':
            if self.illumflat_spat_bsplines is not None:
                return self.illumflat_spat_bsplines
            else:
                msgs.warn("illumflat has no spatial bspline fit - using the pixelflat")
                return self.pixelflat_spat_bsplines
        else:
            if self.pixelflat_spat_bsplines is not None:
                return self.pixelflat_spat_bsplines
            else:
                msgs.warn("pixelflat has no spatial bspline fit - using the illumflat")
                return self.illumflat_spat_bsplines

    def fit2illumflat(self, slits, frametype='illum', initial=False, flexure_shift=None):
        """

        Args:
            slits (:class:`pypeit.slittrace.SlitTraceSet`):
            frametype (str):
                Should the pixel or illum flat spatial profile be generated? (options include: 'illum' or 'pixel').
                The default is to use 'illum' unless frametype='pixel'.
            initial (bool, optional):
            flexure_shift (float, optional):

        Returns:

        """
        illumflat = np.ones(self.shape())
        # Load spatial bsplines
        spat_bsplines = self.get_spat_bsplines(frametype=frametype)

        # Loop
        for slit_idx in range(slits.nslits):
            # Skip masked
            if slits.mask[slit_idx] != 0:
                continue
            # Skip those without a bspline
            # DO it
            _slitid_img = slits.slit_img(slitidx=slit_idx, initial=initial, flexure=flexure_shift)
            onslit = _slitid_img == slits.spat_id[slit_idx]
            spat_coo = slits.spatial_coordinate_image(slitidx=slit_idx,
                                                      initial=initial,
                                                      slitid_img=_slitid_img,
                                                      flexure_shift=flexure_shift)
            illumflat[onslit] = spat_bsplines[slit_idx].value(spat_coo[onslit])[0]
        # TODO -- Update the internal one?  Or remove it altogether??
        return illumflat

    def show(self, frametype='all', slits=None, wcs_match=True):
        """
        Simple wrapper to show_flats()

        Args:
            frametype (str):
                Which flats should be displayed? Must be one of 'illum', 'pixel', or 'all' (default)
            slits:
            wcs_match:

        Returns:

        """
        illumflat_pixel, illumflat_illum = None, None
        # Try to grab the slits
        if slits is None:
            # Warning: This parses the filename, not the Header!
            master_key, master_dir = masterframe.grab_key_mdir(self.filename, from_filename=True)
            try:
                slit_masterframe_name = masterframe.construct_file_name(slittrace.SlitTraceSet, master_key,
                                                                        master_dir=master_dir)
                slits = slittrace.SlitTraceSet.from_file(slit_masterframe_name)
            except:
                msgs.warn('Could not load slits to show with flat-field images. Did you provide the master info??')
        if slits is not None:
            slits.mask_flats(self)
            illumflat_pixel = self.fit2illumflat(slits, frametype='pixel')
            if self.illumflat_spat_bsplines is not None:
                illumflat_illum = self.fit2illumflat(slits, frametype='illum')
        # Decide which frames should be displayed
        if frametype == 'pixel':
            image_list = zip([self.pixelflat_norm, illumflat_pixel, self.pixelflat_raw,
                              self.pixelflat_model, self.pixelflat_spec_illum],
                             ['pixelflat_norm', 'pixelflat_spat_illum', 'pixelflat_raw',
                              'pixelflat_model', 'pixelflat_spec_illum'],
                             [(0.9, 1.1), (0.9, 1.1), None, None,
                              (0.8, 1.2)])
        elif frametype == 'illum':
            image_list = zip([illumflat_illum, self.illumflat_raw],
                             ['illumflat_spat_illum', 'illumflat_raw'],
                             [(0.9, 1.1), None])
        else:
            # Show everything that's available (anything that is None will not be displayed)
            image_list = zip([self.pixelflat_norm, illumflat_pixel, self.pixelflat_raw,
                              self.pixelflat_model, self.pixelflat_spec_illum,
                              illumflat_illum, self.illumflat_raw],
                             ['pixelflat_norm', 'pixelflat_spat_illum', 'pixelflat_raw',
                              'pixelflat_model', 'pixelflat_spec_illum',
                              'illumflat_spat_illum', 'illumflat_raw'],
                             [(0.9, 1.1), (0.9, 1.1), None, None,
                              (0.8, 1.2), (0.9, 1.1), None])
        # Display frames
        show_flats(image_list, wcs_match=wcs_match, slits=slits)



class FlatField(object):
    """
    Builds pixel-level flat-field and the illumination flat-field.

    For the primary methods, see :func:`run`.

    Args:
        rawflatimg (:class:`pypeit.images.pypeitimage.PypeItImage`):
            Processed, combined set of pixelflat images
        spectrograph (:class:`pypeit.spectrographs.spectrograph.Spectrograph`):
            The `Spectrograph` instance that sets the instrument used to
            take the observations.  See usage by
            :class:`pypeit.processimages.ProcessImages` base class.
        flatpar (:class:`pypeit.par.pypeitpar.FlatFieldPar`):
            User-level parameters for constructing the flat-field
            corrections.  If None, the default parameters are used.
        slits (:class:`pypeit.slittrace.SlitTraceSet`):
            The current slit traces.
        wavetilts (:class:`pypeit.wavetilts.WaveTilts`):
            The current wavelength tilt traces; see
        spat_illum_only (boolean):
            Only perform the spatial illumination calculation, and ignore
            the 2D bspline fit. This should only be set to true if you
            want the spatial illumination profile only. If you want to
            simultaneously generate a pixel flat and a spatial
            illumination profile from the same input, this should be
            False (which is the default).

    Attributes:
        rawflatimg (:class:`pypeit.images.pypeitimage.PypeItImage`):
        mspixelflat (`numpy.ndarray`_):
            Normalized flat
        msillumflat (`numpy.ndarray`_):
            Illumination flat
        flat_model (`numpy.ndarray`_):
            Model of the flat
        list_of_spat_bsplines (list):
        spec_illum (`numpy.ndarray`_):
            Image of the relative spectral illumination for a multislit spectrograph

    """

    # Frame type is a class attribute
    frametype = 'pixelflat'
    master_type = 'Flat'


    def __init__(self, rawflatimg, spectrograph, flatpar, slits, wavetilts, wv_calib,
                 spat_illum_only=False):

        # Defaults
        self.spectrograph = spectrograph
        # FieldFlattening parameters
        self.flatpar = flatpar

        # Input data
        self.slits = slits
        self.wavetilts = wavetilts
        self.wv_calib = wv_calib

        # Worth a check
        self.wavetilts.is_synced(self.slits)

        # Attributes unique to this Object
        self.rawflatimg = rawflatimg      # Un-normalized pixel flat as a PypeItImage
        self.mspixelflat = None     # Normalized pixel flat
        self.msillumflat = None     # Illumination flat
        self.flat_model = None      # Model flat
        self.list_of_spat_bsplines = None
        self.spat_illum_only = spat_illum_only
        self.spec_illum = None      # Relative spectral illumination image

        # Completed steps
        self.steps = []

        # Child-specific Internals
#        self.extrap_slit = None
#        self.msblaze = None
#        self.blazeext = None
#        self.slit_profiles = None
#        self.ntckx = None
#        self.ntcky = None

    @property
    def nslits(self):
        """
        Return the number of slits.  Pulled directly from :attr:`slits`, if it exists.
        """
        return 0 if self.slits is None else self.slits.nslits

    # TODO: Need to add functionality to use a different frame for the
    # ilumination flat, e.g. a sky flat
    def run(self, debug=False, show=False):
        """
        Generate normalized pixel and illumination flats.

        This is a simple wrapper for the main flat-field methods:
        
            - Flat-field images are processed using
              :func:`build_pixflat`.
            - Full 2D model, illumination flat, and pixel flat images
              are constructed by :func:`fit`.
            - The results can be shown in a ginga window using
              :func:`show`.

        The method is a simple wrapper for :func:`build_pixflat`,
        :func:`fit`, and :func:`show`.

        Args:
            debug (:obj:`bool`, optional):
                Run in debug mode.
            show (:obj:`bool`, optional):
                Show the results in the ginga viewer.

        Returns:
            :class:`FlatImages`:
        """
        # Fit it
        # NOTE: Tilts do not change and self.slits is updated internally.
        self.fit(spat_illum_only=self.spat_illum_only, debug=debug)

        if show:
            self.show(wcs_match=True)

        # Build the mask
        bpmflats = self.build_mask()

        # Return
        if self.spat_illum_only:
            # Illumination correction only
            return FlatImages(illumflat_raw=self.rawflatimg.image,
                              illumflat_spat_bsplines=np.asarray(self.list_of_spat_bsplines),
                              illumflat_bpm=bpmflats, PYP_SPEC=self.spectrograph.name,
                              spat_id=self.slits.spat_id)
        else:
            # Pixel and illumination correction only
            return FlatImages(pixelflat_raw=self.rawflatimg.image,
                              pixelflat_norm=self.mspixelflat,
                              pixelflat_model=self.flat_model,
                              pixelflat_spat_bsplines=np.asarray(self.list_of_spat_bsplines),
                              pixelflat_bpm=bpmflats, pixelflat_spec_illum=self.spec_illum,
                              PYP_SPEC=self.spectrograph.name, spat_id=self.slits.spat_id)

    def build_mask(self):
        """
        Generate bad pixel mask.

        Returns:
            :obj:`numpy.ndarray` : bad pixel mask
        """
        bpmflats = np.zeros_like(self.slits.mask, dtype=self.slits.bitmask.minimum_dtype())
        for flag in ['SKIPFLATCALIB', 'BADFLATCALIB']:
            bpm = self.slits.bitmask.flagged(self.slits.mask, flag)
            if np.any(bpm):
                bpmflats[bpm] = self.slits.bitmask.turn_on(bpmflats[bpm], flag)
        return bpmflats

    def show(self, wcs_match=True):
        """
        Show all of the flat field products in ginga.

        Args:
            wcs_match (:obj:`bool`, optional):
                Match the WCS of the flat-field images
        """
        # Prepare the images to show, their names and their cuts
        image_list = zip([self.mspixelflat, self.msillumflat, self.rawflatimg.image, self.flat_model],
                         ['pixelflat', 'spat_illum', 'raw', 'model', 'spec_illum'],
                         [(0.9, 1.1), (0.9, 1.1), None, None, (0.8, 1.2)])
        show_flats(image_list, wcs_match=wcs_match, slits=self.slits)

    def fit(self, spat_illum_only=False, debug=False):
        """
        Construct a model of the flat-field image.

        For this method to work, :attr:`rawflatimg` must have been
        previously constructed; see :func:`build_pixflat`.

        The method loops through all slits provided by the :attr:`slits`
        object, except those that have been masked (i.e., slits with
        ``self.slits.mask == True`` are skipped).  For each slit:

            - Collapse the flat-field data spatially using the
              wavelength coordinates provided by the fit to the arc-line
              traces (:class:`pypeit.wavetilts.WaveTilts`), and fit the
              result with a bspline.  This provides the
              spatially-averaged spectral response of the instrument.
              The data used in the fit is trimmed toward the slit
              spatial center via the ``slit_trim`` parameter in
              :attr:`flatpar`.
            - Use the bspline fit to construct and normalize out the
              spectral response.
            - Collapse the normalized flat-field data spatially using a
              coordinate system defined by the left slit edge.  The data
              included in the spatial (illumination) profile calculation
              is expanded beyond the nominal slit edges using the
              ``slit_illum_pad`` parameter in :attr:`flatpar`.  The raw,
              collapsed data is then median filtered (see ``spat_samp``
              in :attr:`flatpar`) and Gaussian filtered; see
              :func:`pypeit.core.flat.illum_filter`.  This creates an
              empirical, highly smoothed representation of the
              illumination profile that is fit with a bspline using
              the :func:`spatial_fit` method.  The
              construction of the empirical illumination profile (i.e.,
              before the bspline fitting) can be done iteratively, where
              each iteration sigma-clips outliers; see the
              ``illum_iter`` and ``illum_rej`` parameters in
              :attr:`flatpar` and
              :func:`pypeit.core.flat.construct_illum_profile`.
            - If requested, the 1D illumination profile is used to
              "tweak" the slit edges by offsetting them to a threshold
              of the illumination peak to either side of the slit center
              (see ``tweak_slits_thresh`` in :attr:`flatpar`), up to a
              maximum allowed shift from the existing slit edge (see
              ``tweak_slits_maxfrac`` in :attr:`flatpar`).  See
              :func:`pypeit.core.tweak_slit_edges`.  If tweaked, the
              :func:`spatial_fit` is repeated to place it on the tweaked
              slits reference frame.
            - Use the bspline fit to construct the 2D illumination image
              (:attr:`msillumflat`) and normalize out the spatial
              response.
            - Fit the residuals of the flat-field data that has been
              independently normalized for its spectral and spatial
              response with a 2D bspline-polynomial fit.  The order of
              the polynomial has been optimized via experimentation; it
              can be changed but you should use extreme caution when
              doing so (see ``twod_fit_npoly``).  The multiplication of
              the 2D spectral response, 2D spatial response, and joint
              2D fit to the high-order residuals define the final flat
              model (:attr:`flat_model`).
            - Finally, the pixel-to-pixel response of the instrument is
              defined as the ratio of the raw flat data to the
              best-fitting flat-field model (:attr:`mspixelflat`)

        This method is the primary method that builds the
        :class:`FlatField` instance, constructing :attr:`mspixelflat`,
        :attr:`msillumflat`, and :attr:`flat_model`.  All of these
        attributes are altered internally.  If the slit edges are to be
        tweaked using the 1D illumination profile (``tweak_slits`` in
        :attr:`flatpar`), the tweaked slit edge arrays in the internal
        :class:`pypeit.edgetrace.SlitTraceSet` object, :attr:`slits`,
        are also altered.

        Used parameters from :attr:`flatpar`
        (:class:`pypeit.par.pypeitpar.FlatFieldPar`) are
        ``spec_samp_fine``, ``spec_samp_coarse``, ``spat_samp``,
        ``tweak_slits``, ``tweak_slits_thresh``,
        ``tweak_slits_maxfrac``, ``rej_sticky``, ``slit_trim``,
        ``slit_illum_pad``, ``illum_iter``, ``illum_rej``, and
        ``twod_fit_npoly``, ``saturated_slits``.

        **Revision History**:

            - 11-Mar-2005  First version written by Scott Burles.
            - 2005-2018    Improved by J. F. Hennawi and J. X. Prochaska
            - 3-Sep-2018 Ported to python by J. F. Hennawi and significantly improved

        Args:
            spat_illum_only (:obj:`bool`, optional):
                If true, only the spatial illumination profile will be calculated.
                The 2D bspline fit will not be performed. This is primarily used
                to build an illumflat.
            debug (:obj:`bool`, optional):
                Show plots useful for debugging. This will block
                further execution of the code until the plot windows
                are closed.

        """
        # TODO: break up this function!  Can it be partitioned into a series of "core" methods?
        # TODO: JFH I wrote all this code and will have to maintain it and I don't want to see it broken up.
        # TODO: JXP This definitely needs breaking up..

        # Initialise with a series of bad splines (for when slits go wrong)
        if self.list_of_spat_bsplines is None:
            self.list_of_spat_bsplines = [bspline.bspline(None) for all in self.slits.spat_id]

        # Set parameters (for convenience;
        spec_samp_fine = self.flatpar['spec_samp_fine']
        spec_samp_coarse = self.flatpar['spec_samp_coarse']
        tweak_slits = self.flatpar['tweak_slits']
        tweak_slits_thresh = self.flatpar['tweak_slits_thresh']
        tweak_slits_maxfrac = self.flatpar['tweak_slits_maxfrac']
        # If sticky, points rejected at each stage (spec, spat, 2d) are
        # propagated to the next stage
        sticky = self.flatpar['rej_sticky']
        trim = self.flatpar['slit_trim']
        pad = self.flatpar['slit_illum_pad']
        # Iteratively construct the illumination profile by rejecting outliers
        npoly = self.flatpar['twod_fit_npoly']
        saturated_slits = self.flatpar['saturated_slits']

        # Build wavelength image -- not always used, but for convenience done here
        slitmask = self.slits.slit_img(initial=True, 
                                       flexure=self.wavetilts.spat_flexure)
        tilts = self.wavetilts.fit2tiltimg(slitmask, 
                                           flexure=self.wavetilts.spat_flexure)
        waveimg = self.wv_calib.build_waveimg(
            tilts, self.slits, spat_flexure=self.wavetilts.spat_flexure)

        # Setup images
        nspec, nspat = self.rawflatimg.image.shape
        rawflat = self.rawflatimg.image
        # Good pixel mask
        gpm = np.ones_like(rawflat, dtype=bool) if self.rawflatimg.bpm is None else (
                1-self.rawflatimg.bpm).astype(bool)

        # Flat-field modeling is done in the log of the counts
        flat_log = np.log(np.fmax(rawflat, 1.0))
        gpm_log = (rawflat > 1.0) & gpm
        # set errors to just be 0.5 in the log
        ivar_log = gpm_log.astype(float)/0.5**2

        # Other setup
        nonlinear_counts = self.spectrograph.nonlinear_counts(self.rawflatimg.detector)

        # TODO -- JFH -- CONFIRM THIS SHOULD BE ON INIT
        # It does need to be *all* of the slits
        median_slit_widths = np.median(self.slits.right_init - self.slits.left_init, axis=0)

        if tweak_slits:
            # NOTE: This copies the input slit edges to a set that can be tweaked.
            self.slits.init_tweaked()

        # TODO: This needs to include a padding check
        # Construct three versions of the slit ID image, all of unmasked slits!
        #   - an image that uses the padding defined by self.slits
        slitid_img_init = self.slits.slit_img(initial=True)
        #   - an image that uses the extra padding defined by
        #     self.flatpar. This was always 5 pixels in the previous
        #     version.
        padded_slitid_img = self.slits.slit_img(initial=True, pad=pad)
        #   - and an image that trims the width of the slit using the
        #     parameter in self.flatpar. This was always 3 pixels in
        #     the previous version.
        # TODO: Fix this for when trim is a tuple
        trimmed_slitid_img = self.slits.slit_img(pad=-trim, initial=True)

        # Prep for results
        self.mspixelflat = np.ones_like(rawflat)
        self.msillumflat = np.ones_like(rawflat)
        self.flat_model = np.zeros_like(rawflat)

        # Allocate work arrays only once
        spec_model = np.ones_like(rawflat)
        norm_spec = np.ones_like(rawflat)
        norm_spec_spat = np.ones_like(rawflat)
        twod_model = np.ones_like(rawflat)
        twod_gpm_out = np.ones_like(rawflat, dtype=np.bool)

        # #################################################
        # Model each slit independently
        for slit_idx, slit_spat in enumerate(self.slits.spat_id):
            # Is this a good slit??
            if self.slits.bitmask.flagged(self.slits.mask[slit_idx], flag=['SHORTSLIT', 'USERIGNORE', 'BADTILTCALIB']):
                msgs.info('Skipping bad slit: {}'.format(slit_spat))
                self.slits.mask[slit_idx] = self.slits.bitmask.turn_on(self.slits.mask[slit_idx], 'BADFLATCALIB')
                continue
            elif self.slits.bitmask.flagged(self.slits.mask[slit_idx], flag=['BOXSLIT']):
                msgs.info('Skipping alignment slit: {}'.format(slit_spat))
                continue
            elif self.slits.bitmask.flagged(self.slits.mask[slit_idx], flag=['BADWVCALIB']) and \
                    (self.flatpar['pixelflat_min_wave'] is not None or self.flatpar['pixelflat_max_wave'] is not None):
                msgs.info('Skipping slit with bad wavecalib: {}'.format(slit_spat))
                self.slits.mask[slit_idx] = self.slits.bitmask.turn_on(self.slits.mask[slit_idx], 'BADFLATCALIB')
                continue

            msgs.info('Modeling the flat-field response for slit spat_id={}: {}/{}'.format(
                        slit_spat, slit_idx+1, self.slits.nslits))

            # Find the pixels on the initial slit
            onslit_init = slitid_img_init == slit_spat

            # Check for saturation of the flat. If there are not enough
            # pixels do not attempt a fit, and continue to the next
            # slit.
            # TODO: set the threshold to a parameter?
            good_frac = np.sum(onslit_init & (rawflat < nonlinear_counts))/np.sum(onslit_init)
            if good_frac < 0.5:
                common_message = 'To change the behavior, use the \'saturated_slits\' parameter ' \
                                 'in the \'flatfield\' parameter group; see here:\n\n' \
                                 'https://pypeit.readthedocs.io/en/latest/pypeit_par.html \n\n' \
                                 'You could also choose to use a different flat-field image ' \
                                 'for this calibration group.'
                if saturated_slits == 'crash':
                    msgs.error('Only {:4.2f}'.format(100*good_frac)
                               + '% of the pixels on slit {0} are not saturated.  '.format(slit_spat)
                               + 'Selected behavior was to crash if this occurred.  '
                               + common_message)
                elif saturated_slits == 'mask':
                    self.slits.mask[slit_idx] = self.slits.bitmask.turn_on(self.slits.mask[slit_idx], 'BADFLATCALIB')
                    msgs.warn('Only {:4.2f}'.format(100*good_frac)
                                                + '% of the pixels on slit {0} are not saturated.  '.format(slit_spat)
                              + 'Selected behavior was to mask this slit and continue with the '
                              + 'remainder of the reduction, meaning no science data will be '
                              + 'extracted from this slit.  ' + common_message)
                elif saturated_slits == 'continue':
                    self.slits.mask[slit_idx] = self.slits.bitmask.turn_on(self.slits.mask[slit_idx], 'SKIPFLATCALIB')
                    msgs.warn('Only {:4.2f}'.format(100*good_frac)
                              + '% of the pixels on slit {0} are not saturated.  '.format(slit_spat)
                              + 'Selected behavior was to simply continue, meaning no '
                              + 'field-flatting correction will be applied to this slit but '
                              + 'pypeit will attempt to extract any objects found on this slit.  '
                              + common_message)
                else:
                    # Should never get here
                    raise NotImplementedError('Unknown behavior for saturated slits: {0}'.format(
                                              saturated_slits))
                continue

            # Demand at least 10 pixels per row (on average) per degree
            # of the polynomial.
            # NOTE: This is not used until the 2D fit. Defined here to
            # be close to the definition of ``onslit``.
            if npoly is None:
                # Approximate number of pixels sampling each spatial pixel
                # for this (original) slit.
                npercol = np.fmax(np.floor(np.sum(onslit_init)/nspec),1.0)
                npoly  = np.clip(7, 1, int(np.ceil(npercol/10.)))
            
            # TODO: Always calculate the optimized `npoly` and warn the
            #  user if npoly is provided but higher than the nominal
            #  calculation?

            # Create an image with the spatial coordinates relative to the left edge of this slit
            spat_coo_init = self.slits.spatial_coordinate_image(slitidx=slit_idx, full=True, initial=True)

            # Find pixels on the padded and trimmed slit coordinates
            onslit_padded = padded_slitid_img == slit_spat
            onslit_trimmed = trimmed_slitid_img == slit_spat

            # ----------------------------------------------------------
            # Collapse the slit spatially and fit the spectral function
            # TODO: Put this stuff in a self.spectral_fit method?

            # Create the tilts image for this slit
            # TODO -- JFH Confirm the sign of this shift is correct!
            _flexure = 0. if self.wavetilts.spat_flexure is None else self.wavetilts.spat_flexure
            tilts = tracewave.fit2tilts(rawflat.shape, self.wavetilts['coeffs'][:,:,slit_idx],
                                        self.wavetilts['func2d'], spat_shift=-1*_flexure)
            # Convert the tilt image to an image with the spectral pixel index
            spec_coo = tilts * (nspec-1)

            # Only include the trimmed set of pixels in the flat-field
            # fit along the spectral direction.
            spec_gpm = onslit_trimmed & gpm_log  # & (rawflat < nonlinear_counts)
            spec_nfit = np.sum(spec_gpm)
            spec_ntot = np.sum(onslit_init)
            msgs.info('Spectral fit of flatfield for {0}/{1} '.format(spec_nfit, spec_ntot)
                      + ' pixels in the slit.')
            # Set this to a parameter?
            if spec_nfit/spec_ntot < 0.5:
                # TODO: Shouldn't this raise an exception or continue to the next slit instead?
                msgs.warn('Spectral fit includes only {:.1f}'.format(100*spec_nfit/spec_ntot)
                          + '% of the pixels on this slit.' + msgs.newline()
                          + '          Either the slit has many bad pixels or the number of '
                            'trimmed pixels is too large.')

            # Sort the pixels by their spectral coordinate.
            # TODO: Include ivar and sorted gpm in outputs?
            spec_gpm, spec_srt, spec_coo_data, spec_flat_data \
                    = flat.sorted_flat_data(flat_log, spec_coo, gpm=spec_gpm)
            # NOTE: By default np.argsort sorts the data over the last
            # axis. Just to avoid the possibility (however unlikely) of
            # spec_coo[spec_gpm] returning an array, all the arrays are
            # explicitly flattened.
            spec_ivar_data = ivar_log[spec_gpm].ravel()[spec_srt]
            spec_gpm_data = gpm_log[spec_gpm].ravel()[spec_srt]

            # Rejection threshold for spectral fit in log(image)
            # TODO: Make this a parameter?
            logrej = 0.5

            # Fit the spectral direction of the blaze.
            # TODO: Figure out how to deal with the fits going crazy at
            #  the edges of the chip in spec direction
            # TODO: Can we add defaults to bspline_profile so that we
            #  don't have to instantiate invvar and profile_basis
            try:
                spec_bspl, spec_gpm_fit, spec_flat_fit, _, exit_status \
                    = fitting.bspline_profile(spec_coo_data, spec_flat_data, spec_ivar_data,
                                            np.ones_like(spec_coo_data), ingpm=spec_gpm_data,
                                            nord=4, upper=logrej, lower=logrej,
                                            kwargs_bspline={'bkspace': spec_samp_fine},
                                            kwargs_reject={'groupbadpix': True, 'maxrej': 5})
            except:
                embed(header='808 of flatfield')

            if exit_status > 1:
                # TODO -- MAKE A FUNCTION
                msgs.warn('Flat-field spectral response bspline fit failed!  Not flat-fielding '
                          'slit {0} and continuing!'.format(slit_spat))
                self.slits.mask[slit_idx] = self.slits.bitmask.turn_on(self.slits.mask[slit_idx], 'BADFLATCALIB')
                continue

            # Debugging/checking spectral fit
            if debug:
                fitting.bspline_qa(spec_coo_data, spec_flat_data, spec_bspl, spec_gpm_fit,
                                 spec_flat_fit, xlabel='Spectral Pixel', ylabel='log(flat counts)',
                                 title='Spectral Fit for slit={:d}'.format(slit_spat))

            if sticky:
                # Add rejected pixels to gpm
                gpm[spec_gpm] = (spec_gpm_fit & spec_gpm_data)[np.argsort(spec_srt)]

            # Construct the model of the flat-field spectral shape
            # including padding on either side of the slit.
            spec_model[...] = 1.
            spec_model[onslit_padded] = np.exp(spec_bspl.value(spec_coo[onslit_padded])[0])
            # ----------------------------------------------------------

            # ----------------------------------------------------------
            # To fit the spatial response, first normalize out the
            # spectral response, and then collapse the slit spectrally.

            # Normalize out the spectral shape of the flat
            norm_spec[...] = 1.
            norm_spec[onslit_padded] = rawflat[onslit_padded] \
                                            / np.fmax(spec_model[onslit_padded],1.0)

            # Find pixels fot fit in the spatial direction:
            #   - Fit pixels in the padded slit that haven't been masked
            #     by the BPM
            spat_gpm = onslit_padded & gpm #& (rawflat < nonlinear_counts)
            #   - Fit pixels with non-zero flux and less than 70% above
            #     the average spectral profile.
            spat_gpm &= (norm_spec > 0.0) & (norm_spec < 1.7)
            #   - Determine maximum counts in median filtered flat
            #     spectrum model.
            spec_interp = interpolate.interp1d(spec_coo_data, spec_flat_fit, kind='linear',
                                               assume_sorted=True, bounds_error=False,
                                               fill_value=-np.inf)
            spec_sm = utils.fast_running_median(np.exp(spec_interp(np.arange(nspec))),
                                                np.fmax(np.ceil(0.10*nspec).astype(int),10))
            #   - Only fit pixels with at least values > 10% of this maximum and no less than 1.
            spat_gpm &= (spec_model > 0.1*np.amax(spec_sm)) & (spec_model > 1.0)

            # Report
            spat_nfit = np.sum(spat_gpm)
            spat_ntot = np.sum(onslit_padded)
            msgs.info('Spatial fit of flatfield for {0}/{1} '.format(spat_nfit, spat_ntot)
                      + ' pixels in the slit.')
            if spat_nfit/spat_ntot < 0.5:
                # TODO: Shouldn't this raise an exception or continue to the next slit instead?
                msgs.warn('Spatial fit includes only {:.1f}'.format(100*spat_nfit/spat_ntot)
                          + '% of the pixels on this slit.' + msgs.newline()
                          + '          Either the slit has many bad pixels, the model of the '
                          'spectral shape is poor, or the illumination profile is very irregular.')

            # First fit -- With initial slits
            exit_status, spat_coo_data,  spat_flat_data, spat_bspl, spat_gpm_fit, \
                spat_flat_fit, spat_flat_data_raw \
                        = self.spatial_fit(norm_spec, spat_coo_init, median_slit_widths[slit_idx],
                                           spat_gpm, gpm, debug=debug)

            if tweak_slits:
                # TODO: Should the tweak be based on the bspline fit?
                # TODO: Will this break if
                left_thresh, left_shift, self.slits.left_tweak[:,slit_idx], right_thresh, \
                    right_shift, self.slits.right_tweak[:,slit_idx] \
                        = flat.tweak_slit_edges(self.slits.left_init[:,slit_idx],
                                                self.slits.right_init[:,slit_idx],
                                                spat_coo_data, spat_flat_data,
                                                thresh=tweak_slits_thresh,
                                                maxfrac=tweak_slits_maxfrac, debug=debug)
                # TODO: Because the padding doesn't consider adjacent
                #  slits, calling slit_img for individual slits can be
                #  different from the result when you construct the
                #  image for all slits. Fix this...

                # Update the onslit mask
                _slitid_img = self.slits.slit_img(slitidx=slit_idx, initial=False)
                onslit_tweak = _slitid_img == slit_spat
                spat_coo_tweak = self.slits.spatial_coordinate_image(slitidx=slit_idx,
                                                               slitid_img=_slitid_img)

                # Construct the empirical illumination profile
                # TODO This is extremely inefficient, because we only need to re-fit the illumflat, but
                #  spatial_fit does both the reconstruction of the illumination function and the bspline fitting.
                #  Only the b-spline fitting needs be reddone with the new tweaked spatial coordinates, so that would
                #  save a ton of runtime. It is not a trivial change becauase the coords are sorted, etc.
                exit_status, spat_coo_data, spat_flat_data, spat_bspl, spat_gpm_fit, \
                    spat_flat_fit, spat_flat_data_raw = self.spatial_fit(
                    norm_spec, spat_coo_tweak, median_slit_widths[slit_idx], spat_gpm, gpm, debug=False)

                spat_coo_final = spat_coo_tweak
            else:
                _slitid_img = slitid_img_init
                spat_coo_final = spat_coo_init
                onslit_tweak = onslit_init

            # Add an approximate pixel axis at the top
            if debug:
                # TODO: Move this into a qa plot that gets saved
                ax = fitting.bspline_qa(spat_coo_data, spat_flat_data, spat_bspl, spat_gpm_fit,
                                      spat_flat_fit, show=False)
                ax.scatter(spat_coo_data, spat_flat_data_raw, marker='.', s=1, zorder=0, color='k',
                           label='raw data')
                # Force the center of the slit to be at the center of the plot for the hline
                ax.set_xlim(-0.1,1.1)
                ax.axvline(0.0, color='lightgreen', linestyle=':', linewidth=2.0,
                           label='original left edge', zorder=8)
                ax.axvline(1.0, color='red', linestyle=':', linewidth=2.0,
                           label='original right edge', zorder=8)
                if tweak_slits and left_shift > 0:
                    label = 'threshold = {:5.2f}'.format(tweak_slits_thresh) \
                                + ' % of max of left illumprofile'
                    ax.axhline(left_thresh, xmax=0.5, color='lightgreen', linewidth=3.0,
                               label=label, zorder=10)
                    ax.axvline(left_shift, color='lightgreen', linestyle='--', linewidth=3.0,
                               label='tweaked left edge', zorder=11)
                if tweak_slits and right_shift > 0:
                    label = 'threshold = {:5.2f}'.format(tweak_slits_thresh) \
                                + ' % of max of right illumprofile'
                    ax.axhline(right_thresh, xmin=0.5, color='red', linewidth=3.0, label=label,
                               zorder=10)
                    ax.axvline(1-right_shift, color='red', linestyle='--', linewidth=3.0,
                               label='tweaked right edge', zorder=20)
                ax.legend()
                ax.set_xlabel('Normalized Slit Position')
                ax.set_ylabel('Normflat Spatial Profile')
                ax.set_title('Illumination Function Fit for slit={:d}'.format(slit_spat))
                plt.show()

            # ----------------------------------------------------------
            # Construct the illumination profile with the tweaked edges
            # of the slit
            if exit_status <= 1:
                # TODO -- JFH -- Check this is ok for flexure!!
                self.msillumflat[onslit_tweak] = spat_bspl.value(spat_coo_final[onslit_tweak])[0]
                self.list_of_spat_bsplines[slit_idx] = spat_bspl
                # No need to proceed further if we just need the illumination profile
                if spat_illum_only:
                    continue
            else:
                # Save the nada
                msgs.warn('Slit illumination profile bspline fit failed!  Spatial profile not '
                          'included in flat-field model for slit {0}!'.format(slit_spat))
                self.slits.mask[slit_idx] = self.slits.bitmask.turn_on(self.slits.mask[slit_idx], 'BADFLATCALIB')
                continue

            # ----------------------------------------------------------
            # Fit the 2D residuals of the 1D spectral and spatial fits.
            msgs.info('Performing 2D illumination + scattered light flat field fit')

            # Construct the spectrally and spatially normalized flat
            norm_spec_spat[...] = 1.
            norm_spec_spat[onslit_tweak] = rawflat[onslit_tweak] / np.fmax(spec_model[onslit_tweak], 1.0) \
                                                    / np.fmax(self.msillumflat[onslit_tweak], 0.01)

            # Sort the pixels by their spectral coordinate. The mask
            # uses the nominal padding defined by the slits object.
            twod_gpm, twod_srt, twod_spec_coo_data, twod_flat_data \
                    = flat.sorted_flat_data(norm_spec_spat, spec_coo, gpm=onslit_tweak)
            # Also apply the sorting to the spatial coordinates
            twod_spat_coo_data = spat_coo_final[twod_gpm].ravel()[twod_srt]
            # TODO: Reset back to origin gpm if sticky is true?
            twod_gpm_data = gpm[twod_gpm].ravel()[twod_srt]
            # Only fit data with less than 30% variations
            # TODO: Make 30% a parameter?
            twod_gpm_data &= np.absolute(twod_flat_data - 1) < 0.3
            # Here we ignore the formal photon counting errors and
            # simply assume that a typical error per pixel. This guess
            # is somewhat aribtrary. We then set the rejection
            # threshold with sigrej_twod
            # TODO: Make twod_sig and twod_sigrej parameters?
            twod_sig = 0.01
            twod_ivar_data = twod_gpm_data.astype(float)/(twod_sig**2)
            twod_sigrej = 4.0

            poly_basis = basis.fpoly(2.0*twod_spat_coo_data - 1.0, npoly)

            # Perform the full 2d fit
            twod_bspl, twod_gpm_fit, twod_flat_fit, _, exit_status \
                    = fitting.bspline_profile(twod_spec_coo_data, twod_flat_data, twod_ivar_data,
                                            poly_basis, ingpm=twod_gpm_data, nord=4,
                                            upper=twod_sigrej, lower=twod_sigrej,
                                            kwargs_bspline={'bkspace': spec_samp_coarse},
                                            kwargs_reject={'groupbadpix': True, 'maxrej': 10})
            if debug:
                # TODO: Make a plot that shows the residuals in the 2D
                # image
                resid = twod_flat_data - twod_flat_fit
                goodpix = twod_gpm_fit & twod_gpm_data
                badpix = np.invert(twod_gpm_fit) & twod_gpm_data

                plt.clf()
                ax = plt.gca()
                ax.plot(twod_spec_coo_data[goodpix], resid[goodpix], color='k', marker='o',
                        markersize=0.2, mfc='k', fillstyle='full', linestyle='None',
                        label='good points')
                ax.plot(twod_spec_coo_data[badpix], resid[badpix], color='red', marker='+',
                        markersize=0.5, mfc='red', fillstyle='full', linestyle='None',
                        label='masked')
                ax.axhline(twod_sigrej*twod_sig, color='lawngreen', linestyle='--',
                           label='rejection thresholds', zorder=10, linewidth=2.0)
                ax.axhline(-twod_sigrej*twod_sig, color='lawngreen', linestyle='--', zorder=10,
                           linewidth=2.0)
#                ax.set_ylim(-0.05, 0.05)
                ax.legend()
                ax.set_xlabel('Spectral Pixel')
                ax.set_ylabel('Residuals from pixelflat 2-d fit')
                ax.set_title('Spectral Residuals for slit={:d}'.format(slit_spat))
                plt.show()

                plt.clf()
                ax = plt.gca()
                ax.plot(twod_spat_coo_data[goodpix], resid[goodpix], color='k', marker='o',
                        markersize=0.2, mfc='k', fillstyle='full', linestyle='None',
                        label='good points')
                ax.plot(twod_spat_coo_data[badpix], resid[badpix], color='red', marker='+',
                        markersize=0.5, mfc='red', fillstyle='full', linestyle='None',
                        label='masked')
                ax.axhline(twod_sigrej*twod_sig, color='lawngreen', linestyle='--',
                           label='rejection thresholds', zorder=10, linewidth=2.0)
                ax.axhline(-twod_sigrej*twod_sig, color='lawngreen', linestyle='--', zorder=10,
                           linewidth=2.0)
#                ax.set_ylim((-0.05, 0.05))
#                ax.set_xlim(-0.02, 1.02)
                ax.legend()
                ax.set_xlabel('Normalized Slit Position')
                ax.set_ylabel('Residuals from pixelflat 2-d fit')
                ax.set_title('Spatial Residuals for slit={:d}'.format(slit_spat))
                plt.show()

            # Save the 2D residual model
            twod_model[...] = 1.
            if exit_status > 1:
                msgs.warn('Two-dimensional fit to flat-field data failed!  No higher order '
                          'flat-field corrections included in model of slit {0}!'.format(slit_spat))
                self.slits.mask[slit_idx] = self.slits.bitmask.turn_on(self.slits.mask[slit_idx], 'BADFLATCALIB')
            else:
                twod_model[twod_gpm] = twod_flat_fit[np.argsort(twod_srt)]
                twod_gpm_out[twod_gpm] = twod_gpm_fit[np.argsort(twod_srt)]


            # Construct the full flat-field model
            # TODO: Why is the 0.05 here for the illumflat compared to the 0.01 above?
            self.flat_model[onslit_tweak] = twod_model[onslit_tweak] \
                                        * np.fmax(self.msillumflat[onslit_tweak], 0.05) \
                                        * np.fmax(spec_model[onslit_tweak], 1.0)

            # Construct the pixel flat
            #trimmed_slitid_img_anew = self.slits.slit_img(pad=-trim, slitidx=slit_idx)
            #onslit_trimmed_anew = trimmed_slitid_img_anew == slit_spat
            self.mspixelflat[onslit_tweak] = rawflat[onslit_tweak]/self.flat_model[onslit_tweak]
            # TODO: Add some code here to treat the edges and places where fits
            #  go bad?

            # Minimum wavelength?
            if self.flatpar['pixelflat_min_wave'] is not None:
                bad_wv = waveimg[onslit_tweak] < self.flatpar['pixelflat_min_wave'] 
                self.mspixelflat[np.where(onslit_tweak)[0][bad_wv]] = 1.
            # Maximum wavelength?
            if self.flatpar['pixelflat_max_wave'] is not None:
                bad_wv = waveimg[onslit_tweak] > self.flatpar['pixelflat_max_wave'] 
                self.mspixelflat[np.where(onslit_tweak)[0][bad_wv]] = 1.

        # No need to continue if we're just doing the spatial illumination
        if spat_illum_only:
            return

        # Set the pixelflat to 1.0 wherever the flat was nonlinear
        self.mspixelflat[rawflat >= nonlinear_counts] = 1.0
        # Set the pixelflat to 1.0 within trim pixels of all the slit edges
        trimmed_slitid_img_new = self.slits.slit_img(pad=-trim, initial=False)
        tweaked_slitid_img = self.slits.slit_img(initial=False)
        self.mspixelflat[(trimmed_slitid_img_new < 0) & (tweaked_slitid_img > 0)] = 1.0

        # Do not apply pixelflat field corrections that are greater than
        # 100% to avoid creating edge effects, etc.
        self.mspixelflat = np.clip(self.mspixelflat, 0.5, 2.0)

        

        # Calculate the relative spectral illumination, if requested
        if self.flatpar['slit_illum_relative']:
            self.spec_illum = self.spectral_illumination(twod_gpm_out, debug=debug)

    def spatial_fit(self, norm_spec, spat_coo, median_slit_width, spat_gpm, gpm, debug=False):
        """
        Perform the spatial fit

        Args:
            norm_spec (`numpy.ndarray`_):
            spat_coo (`numpy.ndarray`_):
                Spatial coordinate array
            median_slit_width (:obj:`float`):
            spat_gpm (`numpy.ndarray`_):
            gpm (`numpy.ndarray`_):
            debug (bool, optional):

        Returns:
            tuple: 7 objects
                 - exit_status (int):
                 - spat_coo_data
                 - spat_flat_data
                 - spat_bspl (:class:`pypeit.bspline.bspline.bspline`): Bspline model of the spatial fit.  Used for illumflat
                 - spat_gpm_fit
                 - spat_flat_fit
                 - spat_flat_data_raw
        """

        # Construct the empirical illumination profile
        _spat_gpm, spat_srt, spat_coo_data, spat_flat_data_raw, spat_flat_data \
            = flat.construct_illum_profile(norm_spec, spat_coo, median_slit_width,
                                           spat_gpm=spat_gpm,
                                           spat_samp=self.flatpar['spat_samp'],
                                           illum_iter=self.flatpar['illum_iter'],
                                           illum_rej=self.flatpar['illum_rej'],
                                           debug=debug)

        if self.flatpar['rej_sticky']:
            # Add rejected pixels to gpm
            gpm[spat_gpm] &= (spat_gpm & _spat_gpm)[spat_gpm]

        # Make sure that the normalized and filtered flat is finite!
        if np.any(np.invert(np.isfinite(spat_flat_data))):
            msgs.error('Inifinities in slit illumination function computation!')

        # Determine the breakpoint spacing from the sampling of the
        # spatial coordinates. Use breakpoints at a spacing of a
        # 1/10th of a pixel, but do not allow a bsp smaller than
        # the typical sampling. Use the bspline class to determine
        # the breakpoints:
        spat_bspl = bspline.bspline(spat_coo_data, nord=4,
                                    bkspace=np.fmax(1.0 / median_slit_width / 10.0,
                                                    1.2 * np.median(np.diff(spat_coo_data))))
        # TODO: Can we add defaults to bspline_profile so that we
        #  don't have to instantiate invvar and profile_basis
        spat_bspl, spat_gpm_fit, spat_flat_fit, _, exit_status \
            = fitting.bspline_profile(spat_coo_data, spat_flat_data,
                                    np.ones_like(spat_flat_data),
                                    np.ones_like(spat_flat_data), nord=4, upper=5.0,
                                    lower=5.0, fullbkpt=spat_bspl.breakpoints)
        # Return
        return exit_status, spat_coo_data, spat_flat_data, spat_bspl, spat_gpm_fit, \
               spat_flat_fit, spat_flat_data_raw

    def spectral_illumination(self, gpm=None, debug=False):
        """
        Generate a relative scaling image for a slit-based IFU. All
        slits are scaled relative to a reference slit, specified in
        the spectrograph settings file.

        Parameters
        ----------
        gpm : `numpy.ndarray`_, None
            Good pixel mask
        debug : bool
            Debug the routine

        Returns
        -------
        scale_model: `numpy.ndarray`_
            An image containing the appropriate scaling
        """
        msgs.info("Deriving spectral illumination profile")
        # Generate a wavelength image
        msgs.info("Generating wavelength image")
        flex = self.wavetilts.spat_flexure
        slitmask = self.slits.slit_img(initial=True, flexure=flex)
        tilts = self.wavetilts.fit2tiltimg(slitmask, flexure=flex)
        waveimg = self.wv_calib.build_waveimg(tilts, self.slits, spat_flexure=flex)
        msgs.info('Performing a joint fit to the flat-field response')
        # Grab some parameters
        trim = self.flatpar['slit_trim']
        rawflat = self.rawflatimg.image.copy() / self.msillumflat.copy()
        # Grab the GPM and the slit images
        if gpm is None:
            gpm = np.ones_like(rawflat, dtype=bool) if self.rawflatimg.bpm is None else (
                    1 - self.rawflatimg.bpm).astype(bool)

        slitid_img_init = self.slits.slit_img(pad=0, initial=True)
        # Obtain the minimum and maximum wavelength of all slits
        mnmx_wv = np.zeros((self.slits.nslits, 2))
        for slit_idx, slit_spat in enumerate(self.slits.spat_id):
            onslit_init = (slitid_img_init == slit_spat)
            mnmx_wv[slit_idx, 0] = np.min(waveimg[onslit_init])
            mnmx_wv[slit_idx, 1] = np.max(waveimg[onslit_init])

        # Obtain relative spectral illumination
        return illum_profile_spectral(rawflat, waveimg, self.slits,
                                      slit_illum_ref_idx=self.flatpar['slit_illum_ref_idx'],
                                      model=None, gpmask=gpm, skymask=None, trim=trim,
                                      flexure=flex)


def show_flats(image_list, wcs_match=True, slits=None):
    """
    Interface to ginga to show a set of flat images

    Args:
        pixelflat (`numpy.ndarray`_):
        illumflat (`numpy.ndarray`_ or None):
        procflat (`numpy.ndarray`_):
        flat_model (`numpy.ndarray`_):
        spec_illum (`numpy.ndarray`_ or None):
        wcs_match (bool, optional):
        slits (:class:`pypeit.slittrace.SlitTraceSet`, optional):

    Returns:

    """
    display.connect_to_ginga(raise_err=True, allow_new=True)
    if slits is not None:
        left, right, mask = slits.select_edges()
        gpm = mask == 0
    # Loop me
    clear = True
    for img, name, cut in image_list:
        if img is None:
            continue
        # TODO: Add an option that shows the relevant stuff in a
        # matplotlib window.
        viewer, ch = display.show_image(img, chname=name, cuts=cut, wcs_match=wcs_match,
                                        clear=clear)
        if slits is not None:
            display.show_slits(viewer, ch, left[:, gpm], right[:, gpm],
                               slit_ids=slits.spat_id[gpm])
        # Turn off clear
        if clear:
            clear = False


def illum_profile_spectral(rawimg, waveimg, slits, slit_illum_ref_idx=0, smooth_npix=None, model=None, gpmask=None, skymask=None, trim=3, flexure=None):
    """
    Determine the relative spectral illumination of all slits.
    Currently only used for image slicer IFUs.

    Parameters
    ----------
    rawimg : `numpy.ndarray`_
        Image data that will be used to estimate the spectral relative sensitivity
    waveimg : `numpy.ndarray`_
        Wavelength image
    slits : :class:`pypeit.slittrace.SlitTraceSet`
        Information stored about the slits
    slit_illum_ref_idx : int
        Index of slit that is used as the reference.
    smooth_npix : int
        smoothing used for determining smoothly varying S/N ratio weights by sn_weights
    model : `numpy.ndarray`_, None
        A model of the rawimg data. If None, rawimg will be used.
    gpmask : `numpy.ndarray`_, None
        Good pixel mask
    skymask : `numpy.ndarray`_, None
        Sky mask
    trim : int
        Number of pixels to trim from the edges of the slit
        when deriving the spectral illumination
    flexure : float, None
        Spatial flexure

    Returns
    -------
    scale_model: `numpy.ndarray`_
        An image containing the appropriate scaling
    """
    msgs.info("Performing relative spectral sensitivity correction (reference slit = {0:d})".format(slit_illum_ref_idx))
    # Setup some helpful parameters
    skymask_now = skymask if (skymask is not None) else np.ones_like(rawimg, dtype=bool)
    gpm = gpmask if (gpmask is not None) else np.ones_like(rawimg, dtype=bool)
    modelimg = model if (model is not None) else rawimg.copy()
    # Setup the slits
    slitid_img_init = slits.slit_img(pad=0, initial=True, flexure=flexure)
    slitid_img_trim = slits.slit_img(pad=-trim, initial=True, flexure=flexure)
    scaleImg = np.ones_like(rawimg)
    modelimg_copy = modelimg.copy()
    # Obtain the minimum and maximum wavelength of all slits
    mnmx_wv = np.zeros((slits.nslits, 2))
    for slit_idx, slit_spat in enumerate(slits.spat_id):
        onslit_init = (slitid_img_init == slit_spat)
        mnmx_wv[slit_idx, 0] = np.min(waveimg[onslit_init])
        mnmx_wv[slit_idx, 1] = np.max(waveimg[onslit_init])
    wavecen = np.mean(mnmx_wv, axis=1)
    # Sort the central wavelengths by those that are closest to the reference slit
    wvsrt = np.argsort(np.abs(wavecen - wavecen[slit_illum_ref_idx]))

    # Prepare wavelength array for all spectra
    dwav = np.max((mnmx_wv[:, 1] - mnmx_wv[:, 0])/slits.nspec)
    numsamp = int((np.max(mnmx_wv) - np.min(mnmx_wv)) / dwav)
    wavebins = np.linspace(np.min(mnmx_wv), np.max(mnmx_wv), numsamp)

    # Start by building a reference spectrum
    onslit_ref_trim = (slitid_img_trim == slits.spat_id[slit_illum_ref_idx]) & gpm & skymask_now
    hist, edge = np.histogram(waveimg[onslit_ref_trim], bins=wavebins, weights=modelimg_copy[onslit_ref_trim])
    cntr, edge = np.histogram(waveimg[onslit_ref_trim], bins=wavebins)
    cntr = cntr.astype(np.float)
    norm = (cntr != 0) / (cntr + (cntr == 0))
    spec_ref = hist * norm
    wave_ref = 0.5 * (wavebins[1:] + wavebins[:-1])

    # Iterate until convergence
    maxiter = 10
    lo_prev, hi_prev = 1.0E-32, 1.0E32
    sn_smooth_npix = smooth_npix if (smooth_npix is not None) else int(np.round(wave_ref.size / 10))
    for rr in range(maxiter):
        # Reset the relative scaling for this iteration
        relscl_model = np.ones_like(rawimg)
        # Build the relative illumination, by successively finding the slits closest in wavelength to the reference
        for ss in range(slits.spat_id.size):
            # Check if this index is the reference
            if wvsrt[ss] == slit_illum_ref_idx: continue
            # Calculate the region of overlap
            onslit_b = (slitid_img_trim == slits.spat_id[wvsrt[ss]])
            onslit_b_init = (slitid_img_init == slits.spat_id[wvsrt[ss]])
            onslit_b_olap = onslit_b & gpm & (waveimg >= mnmx_wv[wvsrt[ss], 0]) & (waveimg <= mnmx_wv[wvsrt[ss], 1]) & skymask_now
            hist, edge = np.histogram(waveimg[onslit_b_olap], bins=wavebins, weights=modelimg_copy[onslit_b_olap])
            cntr, edge = np.histogram(waveimg[onslit_b_olap], bins=wavebins)
            cntr = cntr.astype(np.float)
            cntr *= spec_ref
            norm = (cntr != 0) / (cntr + (cntr == 0))
            arr = hist * norm
            gdmask = (arr != 0)
            # Calculate a smooth version of the relative response
            relscale = coadd.smooth_weights(arr, gdmask, sn_smooth_npix)
            rescale_model = interpolate.interp1d(wave_ref, relscale, kind='linear', bounds_error=False,
                                                 fill_value="extrapolate")(waveimg[onslit_b_init])
            # Store the result
            relscl_model[onslit_b_init] = rescale_model.copy()

            # Build a new reference spectrum to increase wavelength coverage of the reference spectrum (and improve S/N)
            onslit_ref_trim = onslit_ref_trim | (onslit_b & gpm & skymask_now)
            hist, edge = np.histogram(waveimg[onslit_ref_trim], bins=wavebins, weights=modelimg_copy[onslit_ref_trim]/relscl_model[onslit_ref_trim])
            cntr, edge = np.histogram(waveimg[onslit_ref_trim], bins=wavebins)
            cntr = cntr.astype(np.float)
            norm = (cntr != 0) / (cntr + (cntr == 0))
            spec_ref = hist * norm
        minv, maxv = np.min(relscl_model), np.max(relscl_model)
        if 1/minv + maxv > lo_prev+hi_prev:
            # Adding noise, so break
            # NOTE : The best precision one might hope for is about:
            # 1.4826 * MAD(arr) / np.sqrt(sn_smooth_npix/ 10)  # /10 comes from the coadd.smooth_weights function
            break
        else:
            lo_prev, hi_prev = 1/minv, maxv
        msgs.info("Iteration {0:d} :: Minimum/Maximum scales = {1:.5f}, {2:.5f}".format(rr + 1, minv, maxv))
        # Store rescaling
        scaleImg *= relscl_model
        #rawimg_copy /= relscl_model
        modelimg_copy /= relscl_model
        if max(abs(1/minv), abs(maxv)) < 1.001:  # Relative accruacy of 0.1% is sufficient
            break
    debug = False
    if debug:
        ricp = rawimg.copy()
        for ss in range(slits.spat_id.size):
            onslit_ref_trim = (slitid_img_trim == slits.spat_id[ss]) & gpm & skymask_now
            hist, edge = np.histogram(waveimg[onslit_ref_trim], bins=wavebins, weights=ricp[onslit_ref_trim]/scaleImg[onslit_ref_trim])
            histScl, edge = np.histogram(waveimg[onslit_ref_trim], bins=wavebins, weights=scaleImg[onslit_ref_trim])
            histAlt, edge = np.histogram(waveimg[onslit_ref_trim], bins=wavebins, weights=scaleImgAlt[onslit_ref_trim])
            cntr, edge = np.histogram(waveimg[onslit_ref_trim], bins=wavebins)
            cntr = cntr.astype(np.float)
            norm = (cntr != 0) / (cntr + (cntr == 0))
            spec_ref = hist * norm
            scale_ref = histScl * norm
            scale_refAlt = histAlt * norm
            #plt.subplot(211)
            plt.plot(wave_ref, scale_refAlt)
            #plt.subplot(212)
            plt.plot(wave_ref, scale_ref)
        plt.show()

    return scaleImg


def merge(init_cls, merge_cls):
    """
    Merge merge_cls into init_cls, and return a merged :class:`pypeit.flatfield.FlatImages` class.
    If an element exists in both init_cls and merge_cls, the merge_cls value is taken

    Parameters
    ----------
    init_cls : :class:`pypeit.flatfield.FlatImages`
        Initial class (the elements of this class will be considered the default)
    merge_cls : :class:`pypeit.flatfield.FlatImages`
        The non-zero elements will be merged into init_cls.

    Returns
    -------
    :class:`pypeit.flatfield.FlatImages` : A new instance of the FlatImages class with merged properties.
    """
    # Check the class to be merged in is not None
    if merge_cls is None:
        return init_cls
    # Initialise variables
    # extract all elements that are prefixed with 'pixelflat_' or 'illumflat_'
    keys = [a for a in list(init_cls.__dict__.keys()) if '_' in a and a.split('_')[0] in ['illumflat', 'pixelflat']]
    dd = dict()
    for key in keys:
        dd[key] = None
    # Cherry pick the values from each class
    dd['PYP_SPEC'] = merge_cls.PYP_SPEC if init_cls.PYP_SPEC is None else init_cls.PYP_SPEC
    dd['spat_id'] = merge_cls.spat_id if init_cls.spat_id is None else init_cls.spat_id
    for key in keys:
        mrg = False
        val = None
        namespace = dict({'val': val, 'init_cls':init_cls, 'merge_cls':merge_cls, 'mrg':mrg})
        exec("val = init_cls.{0:s}".format(key), namespace)
        exec("mrg = merge_cls.{0:s} is not None".format(key), namespace)
        if namespace['mrg']:
            exec("val = merge_cls.{0:s}".format(key), namespace)
        dd[key] = namespace['val']
    # Construct the merged class
    return FlatImages(**dd)






"""
Module to run tests on FlatField class
Requires files in Development suite and an Environmental variable
"""
import os

import pytest
import glob
import shutil

import numpy as np

from pypeit import calibrations
from pypeit.par import pypeitpar
from pypeit.spectrographs.util import load_spectrograph
from pypeit import wavecalib
from IPython import embed

from pypeit.tests.tstutils import dev_suite_required, dummy_fitstbl

def data_path(filename):
    data_dir = os.path.join(os.path.dirname(__file__), 'files')
    return os.path.join(data_dir, filename)


@pytest.fixture
def fitstbl():
    if os.getenv('PYPEIT_DEV') is None:
        fitstbl = dummy_fitstbl(directory=data_path(''))
        fitstbl['framebit'][0] = fitstbl.type_bitmask.turn_off(fitstbl['framebit'][0], flag='bias')
        fitstbl['filename'][1] = 'b1.fits.gz'
        fitstbl['filename'][5] = 'b27.fits.gz'
        return fitstbl

    fitstbl = dummy_fitstbl(directory=os.path.join(os.getenv('PYPEIT_DEV'), 'RAW_DATA',
                                                   'shane_kast_blue', '600_4310_d55'))
    # Set the Bias to known
    fitstbl['framebit'][0] = fitstbl.type_bitmask.turn_off(fitstbl['framebit'][0], flag='bias')
    fitstbl['filename'][1] = 'b1.fits.gz'
    for ii in range(2,5):
        fitstbl['filename'][ii] = 'b{0}.fits.gz'.format(ii)
    fitstbl['filename'][5] = 'b27.fits.gz'

    return fitstbl


@pytest.fixture
def multi_caliBrate(fitstbl):
    # Grab a science file for configuration specific parameters
    for idx, row in enumerate(fitstbl):
        if 'science' in row['frametype']:
            sci_file = os.path.join(row['directory'], row['filename'])
            break
    # Par
    spectrograph = load_spectrograph('shane_kast_blue')
    par = spectrograph.config_specific_par(sci_file)
    turn_off = dict(use_biasimage=False)
    par.reset_all_processimages_par(**turn_off)
    #
    calib_par = par['calibrations']
    calib_par['bpm_usebias'] = False
    #calib_par['biasframe']['useframe'] = 'none' # Only use overscan
    calib_par['slitedges']['sync_predict'] = 'nearest'

    multi_caliBrate = calibrations.MultiSlitCalibrations(fitstbl, calib_par, spectrograph,
                                                         data_path('Masters'))
    return reset_calib(multi_caliBrate)


def reset_calib(calib):
    # Find the first science row
    frame = calib.fitstbl.find_frames('science', index=True)[0]
    # Set
    det = 1
    calib.set_config(frame, det)
    return calib


###################################################
# TESTS BEGIN HERE

def test_instantiate(fitstbl):
    par = pypeitpar.PypeItPar()
    spectrograph = load_spectrograph('shane_kast_blue')
    caliBrate = calibrations.MultiSlitCalibrations(fitstbl, par['calibrations'], spectrograph,
                                                   data_path('Masters'))


def test_bias(multi_caliBrate):
    """
    This should produce nothing as we have no bias frames

    Returns:

    """
    multi_caliBrate.get_bias()


def test_arc(multi_caliBrate):
    arc = multi_caliBrate.get_arc()
    assert arc.image.shape == (2048,350)

    # Cleanup
    shutil.rmtree(multi_caliBrate.master_dir)


def test_tiltimg(multi_caliBrate):
    tilt = multi_caliBrate.get_tiltimg()
    assert tilt.image.shape == (2048,350)

    # Cleanup
    shutil.rmtree(multi_caliBrate.master_dir)

def test_bpm(multi_caliBrate):
    # Prep
    multi_caliBrate.shape = (2048,350)
    # Build
    bpm = multi_caliBrate.get_bpm()
    assert bpm.shape == (2048,350)
    assert np.sum(bpm) == 0.


@dev_suite_required
def test_it_all(multi_caliBrate):
    # Setup
    multi_caliBrate.shape = (2048,350)
    #multi_caliBrate.get_pixlocn()
    multi_caliBrate.get_bpm()
    multi_caliBrate.get_arc()
    multi_caliBrate.get_tiltimg()
    slits = multi_caliBrate.get_slits()
    assert slits.PYP_SPEC == 'shane_kast_blue', 'Wrong spectrograph'
    assert (slits.nspec, slits.nspat) == multi_caliBrate.shape, 'Wrong image shape'
    assert slits.nslits == 1, 'Incorrect number of slits'
    assert slits.left_init.shape == (2048,1), 'Incorrect shape for left'
    assert slits.left_tweak is None, 'Tweaks should not exist'

    wv_calib = multi_caliBrate.get_wv_calib()
    assert isinstance(wv_calib, wavecalib.WaveCalib)
    assert 175 in wv_calib.spat_ids
    assert wv_calib.wv_fits[0]['rms'] < 0.2

    waveTilts = multi_caliBrate.get_tilts()
    assert waveTilts.nslit == 1

    multi_caliBrate.get_flats()
    flatImages = multi_caliBrate.get_flats()
    assert flatImages.pixelflat_norm.shape == (2048,350)
    assert flatImages.fit2illumflat(slits).shape == (2048,350)

    # Wave image
    slitmask = slits.slit_img()
    tilts = waveTilts.fit2tiltimg(slitmask)

    #
    mswave = wv_calib.build_waveimg(tilts, slits)
    assert mswave.shape == (2048,350)

@dev_suite_required
def test_reuse(multi_caliBrate, fitstbl):
    """
    Test that Calibrations appropriately reuses existing calibrations frames.
    """
    # In case of previous data or failures
    if os.path.isdir(multi_caliBrate.master_dir):
        shutil.rmtree(multi_caliBrate.master_dir)

    os.makedirs(multi_caliBrate.master_dir)

    # Perform the calibrations and check that the data are correctly
    # stored in memory
    multi_caliBrate.shape = (2048,350)
    multi_caliBrate.get_bpm()

#   Make them all
#    assert list(multi_caliBrate_reuse.calib_dict.keys()) == ['A_1_01'], 'Incorrect master key'
#    assert list(multi_caliBrate_reuse.calib_dict['A_1_01'].keys()) == ['bpm'], \
#                'Incorrect list of master types in memory'
    msarc = multi_caliBrate.get_arc()
#    assert list(multi_caliBrate_reuse.calib_dict['A_1_01'].keys()) == ['bpm', 'arc'], \
#                'Incorrect list of master types in memory'
    msarc = multi_caliBrate.get_tiltimg()
#    assert list(multi_caliBrate_reuse.calib_dict['A_1_01'].keys()) == ['bpm', 'arc', 'tiltimg'], \
#                'Incorrect list of master types in memory'
    multi_caliBrate.get_slits()
#    assert list(multi_caliBrate_reuse.calib_dict['A_1_01'].keys()) == ['bpm', 'arc', 'tiltimg', 'trace'], \
#                'Incorrect list of master types in memory'
    multi_caliBrate.get_wv_calib()
#    assert list(multi_caliBrate_reuse.calib_dict['A_1_01'].keys()) \
#                == ['bpm', 'arc', 'tiltimg','trace', 'wavecalib'], \
#                'Incorrect list of master types in memory'
    multi_caliBrate.get_tilts()
#    assert list(multi_caliBrate_reuse.calib_dict['A_1_01'].keys()) \
#                == ['bpm', 'arc', 'tiltimg', 'trace', 'wavecalib', 'wavetilts'], \
#                'Incorrect list of master types in memory'
    multi_caliBrate.get_flats()
#    assert list(multi_caliBrate_reuse.calib_dict['A_1_01'].keys()) \
#                == ['bpm', 'arc', 'tiltimg', 'trace', 'wavecalib', 'wavetilts',
#                    'flatimages'], \
#                'Incorrect list of master types in memory'

    # Reset
    #reset_calib(multi_caliBrate_reuse)
    spectrograph = load_spectrograph('shane_kast_blue')
    par = spectrograph.default_pypeit_par()
    multi_caliBrate_reuse = calibrations.MultiSlitCalibrations(fitstbl, par['calibrations'],
                                                               spectrograph, data_path('Masters'))
    multi_caliBrate_reuse.reuse_masters = True
    reset_calib(multi_caliBrate_reuse)

    # Read the calibrations
    #   - These don't source a master file
    multi_caliBrate_reuse.shape = (2048,350)
    multi_caliBrate_reuse.get_bpm()
    #   - The arc is the first sourced master
    assert 'arc' not in multi_caliBrate_reuse.master_key_dict.keys(), \
            'arc master key should not be defined yet'
    _msarc = multi_caliBrate_reuse.get_arc()
    assert multi_caliBrate_reuse.msarc is not None, 'Should find cached data.'
    _msarc = multi_caliBrate_reuse.get_tiltimg()
    assert multi_caliBrate_reuse.mstilt is not None, 'Should find cached data.'
    # JXP took these out of the class
    #assert os.path.isfile(multi_caliBrate_reuse.arcImage.master_file_path), \
    #        'Should find master file.'
    #assert multi_caliBrate_reuse.arcImage.load() is not None, \
    #        'Load should not return None'
    #   - Make sure the rest of the steps complete
    multi_caliBrate_reuse.get_slits()
    multi_caliBrate_reuse.get_wv_calib()
    multi_caliBrate_reuse.get_tilts()
    multi_caliBrate_reuse.get_flats()

    # Clean-up
    shutil.rmtree(multi_caliBrate_reuse.master_dir)




""" Object to load and process a single raw image

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""

import os
import inspect
from copy import deepcopy

from IPython import embed

import numpy as np

from astropy import stats

from pypeit import msgs
from pypeit.core import parse
from pypeit.core import procimg
from pypeit.core import flat
from pypeit.core import flexure
from pypeit.images import pypeitimage
from pypeit import utils
from pypeit.display import display

# TODO: I don't understand why we have some of these attributes.  E.g., why do
# we need both hdu and headarr?
class RawImage:
    """
    Class to load and process a raw image

    Args:
        ifile (:obj:`str`):
            File with the data.
        spectrograph (:class:`~pypeit.spectrographs.spectrograph.Spectrograph`):
            The spectrograph from which the data was collected.
        det (:obj:`int`):
            The detector to load/process.

    Attributes:
        filename (:obj:`str`):
            Original file name with the data.
        spectrograph (:class:`~pypeit.spectrograph.spectrographs.Spectrograph`):
            Spectrograph instance with the instrument-specific properties and
            methods.
        det (:obj:`int`):
            1-indexed detector number
        detector (:class:`~pypeit.images.detector_container.DetectorContainer`):
            Detector characteristics
        rawimage (`numpy.ndarray`_):
            The raw, not trimmed or reoriented, image data for the detector.
        hdu (`astropy.io.fits.HDUList`_):
            The full list of HDUs provided by :attr:`filename`.
        exptime (:obj:`float`):
            Frame exposure time in seconds.
        rawdatasec_img (`numpy.ndarray`_):
            The original, not trimmed or reoriented, image identifying which
            amplifier was used to read each section of the raw image.
        oscansec_img (`numpy.ndarray`_):
            The original, not trimmed or reoriented, image identifying the
            overscan regions in the raw image read by each amplifier.
        headarr (:obj:`list`):
            A list of `astropy.io.fits.Header`_ objects with the headers for all
            extensions in :attr:`hdu`.
        image (`numpy.ndarray`_):
            The processed image.  This starts as identical to :attr:`rawimage`
            and then altered by the processing steps; see :func:`process`.
        ronoise (:obj:`list`):
            The readnoise (in e-/ADU) for each of the detector amplifiers.
        par (:class:`~pypeit.par.pypeitpar.ProcessImagesPar`):
            Parameters that dictate the processing of the images.
        ivar (`numpy.ndarray`_):
            The inverse variance of :attr:`image`, the processed image.
        rn2img (`numpy.ndarray`_):
            The readnoise variance image.
        proc_var (`numpy.ndarray`_):
            The sum of the variance components added by the image processing;
            i.e., the error in the overscan subtraction, bias subtraction, etc.
        base_var (`numpy.ndarray`_):
            The base-level variance in the processed image.  See
            :func:`~pypeit.core.procimg.base_variance`.
        var (`numpy.ndarray`_):
            The aggregate variance in the processed image.  This is the primary
            array used during :func:`process` to track uncertainties;
            :attr:`ivar` is created by inverting this at the end of the
            processing method.
        steps (:obj:`dict`):
            Dictionary containing a set of booleans that track the processing
            steps that have been performed.
        datasec_img (`numpy.ndarray`_):
            Image identifying which amplifier was used to read each section of
            the *processed* image.
        spat_flexure_shift (:obj:`float`):
            The spatial flexure shift in pixels, if calculated
    """
    def __init__(self, ifile, spectrograph, det):

        # Required parameters
        self.filename = ifile
        self.spectrograph = spectrograph
        self.det = det

        # Load the raw image and the other items of interest
        self.detector, self.rawimage, self.hdu, self.exptime, self.rawdatasec_img, \
                self.oscansec_img = self.spectrograph.get_rawimage(self.filename, self.det)

        # Grab items from rawImage (for convenience and for processing)
        #   Could just keep rawImage in the object, if preferred
        self.headarr = deepcopy(self.spectrograph.get_headarr(self.hdu))

        # Key attributes
        self.image = self.rawimage.copy()
        self.datasec_img = self.rawdatasec_img.copy()
        # NOTE: Prevent estimate_readnoise() from altering self.detector using
        # deepcopy
        self.ronoise = deepcopy(self.detector['ronoise'])

        # Attributes
        self.par = None
        self.ivar = None
        self.rn2img = None
        self.proc_var = None
        self.base_var = None
        self.spat_flexure_shift = None
        self.img_scale = None
        self._bpm = None

        # All possible processing steps.  NOTE: These have to match the
        # method/function names.  Their order here matches there execution order
        # in self.process(), but that's not necessary.
        self.steps = dict(apply_gain=False,
                          subtract_pattern=False,
                          subtract_overscan=False,
                          trim=False,
                          orient=False,
                          subtract_bias=False,
                          subtract_dark=False,
                          spatial_flexure_shift=False,
                          flatfield=False)

    @property
    def bpm(self):
        """
        Generate and return the bad pixel mask for this image.

        .. warning::

            BPMs are for processed (e.g. trimmed, rotated) images only!

        Returns:
            `numpy.ndarray`_:  Bad pixel mask with a bad pixel = 1

        """
        if self._bpm is None:
            # TODO: Was this ever being called?
#            self._bpm = self.spectrograph.bpm(shape=self.image.shape, filename=self.filename,
#                                              det=self.det)
            # TODO: Pass bias if there is one?
            self._bpm = self.spectrograph.bpm(self.filename, self.det, shape=self.image.shape)
        return self._bpm

    @property
    def use_flat(self):
        """
        Return a flag setting if the flat data should be used in the image
        processing.
        """
        if self.par is None:
            return False
        # TODO: At the moment, we can only perform any of the flat-field
        # corrections if we are applying the pixel-flat correction.
#        return self.par['use_pixelflat'] or self.par['use_specillum'] or self.par['use_illumflat']
        return self.par['use_pixelflat']

    def apply_gain(self, force=False):
        """
        Use the gain to convert images from ADUs to electrons/counts.

        Conversion applied to :attr:`image, :attr:`var`, and :attr:`rn2img`.

        Args:
            force (:obj:`bool`, optional):
                Force the gain to be applied to the image, even if the step log
                (:attr:`steps`) indicates that it already has been.
        """
        step = inspect.stack()[0][3]
        if self.steps[step] and not force:
            # Already applied
            msgs.warn('Gain was already applied.')
            return

        # Behavior needs to be different if the image has been trimmed or not
        gain = procimg.gain_frame(self.datasec_img, np.atleast_1d(self.detector['gain']))
        # TODO: Could also check if steps['trim'] is true.  Is either better or worse?
        if self.rawimage.shape is not None and self.image.shape == self.rawimage.shape:
            # Image is raw, so need to include overscan sections
            # TODO: This only works assuming that there is *no* overlap between
            # oscansec_img and datasec_img.  There shouldn't be, but the code
            # doesn't check this...
            gain += procimg.gain_frame(self.oscansec_img, np.atleast_1d(self.detector['gain']))
        self.image *= gain
        # NOTE: In ``process``, ``apply_gain`` is called first, meaning that all
        # the variance arrays should be None.
        self.steps[step] = True

    def build_ivar(self):
        """
        Generate the inverse variance in the image.

        This is a simple wrapper for :func:`~pypeit.core.procimg.base_variance`
        and :func:`~pypeit.core.procimg.variance_model`.

        Returns:
            `numpy.ndarray`_: The inverse variance in the image.
        """
        # TODO: This approach assumes that the dark-current provided/known for
        # each detector is more accurate and precise than using the counts in
        # the dark image itself (assuming one is available) to calculate the
        # dark-current shot noise.  Instead, we could pass an image of the dark
        # current, if available.
        npix = np.prod(parse.parse_binning(self.detector.binning))
        darkcurr = npix*self.detector['darkcurr'] \
                    if self.detector['darkcurr'] > 0 and self.par['shot_noise'] else None
        self.base_var = procimg.base_variance(self.rn2img, darkcurr=darkcurr, exptime=self.exptime,
                                              proc_var=self.proc_var, count_scale=self.img_scale)
        var = procimg.variance_model(self.base_var,
                                     counts=self.image if self.par['shot_noise'] else None,
                                     count_scale=self.img_scale,
                                     noise_floor=self.par['noise_floor'])
        return utils.inverse(var)

    def estimate_readnoise(self):
        """
        Estimate the readnoise (in electrons) based on the overscan regions of
        the image.

        If the readnoise is not known for any of the amplifiers (i.e., if
        :attr:`ronoise` is :math:`\leq 0`) or if explicitly requested using the
        ``empirical_rn`` parameter, the function estimates it using the standard
        deviation in the overscan region.

        .. warning::

            This function edits :attr:`ronoise` in place.
        """
        if self.oscansec_img.shape != self.image.shape:
            msgs.error('Must estimate readnoise before trimming the image.')
        for amp in range(len(self.ronoise)):
            if self.ronoise[amp] > 0 and not self.par['empirical_rn']:
                # Skip if the readnoise is defined and the empirical readnoise
                # estimate was not explicitly requested.
                continue
            if not np.any(self.oscansec_img==amp+1):
                msgs.error(f'Cannot estimate readnoise for amplifier {amp+1}; no overscan region!')
            gain = 1. if self.steps['apply_gain'] else self.detector['gain'][amp]
            biaspix = self.image[self.oscansec_img==amp+1] * gain
            self.ronoise[amp] = stats.sigma_clipped_stats(biaspix, sigma=5)[-1]
            msgs.info(f'Estimated readnoise of amplifier {amp+1} = {self.ronoise[amp]:.3f} e-')

    def build_rn2img(self, units='e-', digitization=False):
        """
        Generate the model readnoise variance image (:attr:`rn2img`).

        This is primarily a wrapper for :func:`~pypeit.core.procimg.rn2_frame`.

        Args:
            units (:obj:`str`, optional):
                Units for the output variance.  Options are ``'e-'`` for
                variance in square electrons (counts) or ``'ADU'`` for square
                ADU.
            digitization (:obj:`bool`, optional):
                Include digitization error in the calculation.

        Returns:
            `numpy.ndarray`_: Readnoise variance image.
        """
        if not np.all(self.ronoise > 0):
            # TODO: Consider just calling estimate_readnoise here...
            msgs.error('Some readnoise values <=0; first call estimate_readnoise.')

        # Compute and return the readnoise variance image 
        rn2 = procimg.rn2_frame(self.datasec_img, self.ronoise, units=units,
                                gain=self.detector['gain'], digitization=digitization)
        # TODO: Could also check if steps['trim'] is true.  Is either better or worse?
        if self.rawimage.shape is not None and self.image.shape == self.rawimage.shape \
                and np.any(self.oscansec_img > 0):
            # Image is raw, so need to include overscan sections
            # TODO: This only works assuming that there is *no* overlap between
            # oscansec_img and datasec_img.  There shouldn't be, but the code
            # doesn't check this...
            rn2 += procimg.rn2_frame(self.oscansec_img, self.ronoise, units=units,
                                     gain=self.detector['gain'], digitization=digitization)
        return rn2

    def process(self, par, bpm=None, flatimages=None, bias=None, slits=None, debug=False,
                dark=None):
        """
        Process the image.

        See further discussion of :ref:`image_proc` in ``PypeIt``.

        .. note::

            The processing step order is currently 'frozen' as is.  However, in
            the future, we may choose to allow optional ordering.

        The processing steps used (depending on the parameter toggling in
        :attr:`par`), in the order they will be applied are:

            #. :func:`apply_gain`: The first step is to convert the image units
               from ADU to electrons, amp by amp, using the gain provided by the
               :class:`~pypeit.images.detector_container.DetectorContainer`
               instance(s) for each
               :class:`~pypeit.spectrographs.spectrograph.Spectrograph`
               subclass.

            #. :func:`subtract_pattern`: Analyze and subtract sinusoidal pattern
               noise from the image; see
               :func:`~pypeit.core.procimg.subtract_pattern`.

            #. :func:`build_rn2img`: Construct the readnoise variance image,
               which includes readnoise and digitization error.  If any of the
               amplifiers on the detector do not have a measured readnoise or if
               explicitly requested using the ``empirical_rn`` parameter, the
               readnoise is estimated using :func:`estimate_readnoise`.
            
            #. :func:`subtract_overscan`: Use the detector overscan region to
               measure and subtract the frame-dependent bias level along the
               readout direction.

            #. :func:`trim`: Trim the image to include the data regions only
               (i.e. remove the overscan).

            #. :func:`orient`: Orient the image in the PypeIt orientation ---
               spectral coordinates ordered along the first axis and spatial
               coordinates ordered along the second, ``(spec, spat)`` --- with
               blue to red going from small pixel numbers to large.

            #. :func:`subtract_bias`: Subtract the master bias image.  The shape
               and orientation of the bias image must match the *processed*
               image.  I.e., if you trim and orient this image, you must also
               have trimmed and oriented the bias frames.

            #. :func:`subtract_dark`: Subtract the master dark image.  The shape
               and orientation of the dark image must match the *processed*
               image.  I.e., if you trim and orient this image, you must also
               have trimmed and oriented the dark frames.  The dark image is
               *automatically* scaled by the ratio of the exposure times to
               ensure the counts/s in the dark are removed from the image being
               processed.

            #. :func:`spatial_flexure_shift`: Measure any spatial shift due to
               flexure.

            #. :func:`flatfield`: Divide by the pixel-to-pixel, spatial and
               spectral response functions.

            #. :func:`build_ivar`: Construct a model estimate of the variance in
               the image based in the readnoise, errors from the additive
               processing steps, shot-noise from the observed counts (see the
               ``shot_noise`` parameter), a rescaling due to the flat-field
               correction, and a noise floor that sets a maximum S/N per pixel
               (see the ``noise_floor`` parameter); see
               :func:`~pypeit.core.procimg.variance_model`.
            
            #. :func:`~pypeit.images.pypeitimage.PypeItImage.build_crmask`:
               Generate a cosmic-ray mask

        Args:
            par (:class:`~pypeit.par.pypeitpar.ProcessImagesPar`):
                Parameters that dictate the processing of the images.  See
                :class:`pypeit.par.pypeitpar.ProcessImagesPar` for the
                defaults.
            bpm (`numpy.ndarray`_, optional):
                The bad-pixel mask.  This is used to *overwrite* the default
                bad-pixel mask for this spectrograph.  The shape must match a
                trimmed and oriented processed image.
            flatimages (:class:`~pypeit.flatfield.FlatImages`, optional):
                Flat-field images used to apply flat-field corrections.
            bias (:class:`~pypeit.images.pypeitimage.PypeItImage`, optional):
                Bias image for bias subtraction.
            slits (:class:`~pypeit.slittrace.SlitTraceSet`, optional):
                Used to calculate spatial flexure between the image and the
                slits, if requested via the ``spat_flexure_correct`` parameter
                in :attr:`par`; see
                :func:`~pypeit.core.flexure.spat_flexure_shift`.  Also used to
                construct the slit illumination profile, if requested via the
                ``use_illumflat`` parameter in :attr:`par`; see
                :func:`~pypeit.flatfield.FlatImages.fit2illumflat`.
            debug (:obj:`bool`, optional):
                Run in debug mode.
            dark (:class:`~pypeit.images.pypeitimage.PypeItImage`):
                Dark image

        Returns:
            :class:`~pypeit.images.pypeitimage.PypeItImage`: The processed
            image.
        """
        # TODO: Reset steps to all be false at the beginning of the function?
        # If we're careful with book-keeping of the attributes (i.e., resetting
        # image to rawimage, etc), calling process multiple times could be a way
        # for developers to *re*process an image multiple times from scratch
        # with different parameters to test changes.

        # Set input to attributes
        self.par = par
        if bpm is not None:
            self._bpm = bpm

        # Check the input
        if self.par['use_biasimage'] and bias is None:
            msgs.error('No bias available for bias subtraction!')
        if self.par['use_darkimage'] and dark is None:
            msgs.error('No dark available for dark subtraction!')
        if self.par['spat_flexure_correct'] and slits is None:
            msgs.error('Spatial flexure correction requested but no slits provided.')
        if self.use_flat and flatimages is None:
            msgs.error('Flat-field corrections requested but no flat-field images generated '
                       'or provided.  Make sure you have flat-field images in your PypeIt file!')

        #   - Convert from ADU to electron counts.
        if self.par['apply_gain']:
            self.apply_gain()

        # Apply additive corrections.  The order matters and is fixed.
        #
        #   - Subtract any fixed pattern defined for the instrument.  NOTE: This
        #     step *must* be done before subtract_overscan
        if self.par['use_pattern']:
            self.subtract_pattern()

        #   - Estimate the readnoise using the overscan regions.  NOTE: This
        #     needs to be done *after* any pattern subtraction.
        if self.par['empirical_rn'] or not np.all(self.ronoise > 0):
            self.estimate_readnoise()

        #   - Initialize the variance images.  Starts with the shape and
        #     orientation of the raw image.  Note that the readnoise variance
        #     and "process" variance images are kept separate because they are
        #     used again when modeling the noise during object extraction.
        self.rn2img = self.build_rn2img()
        self.proc_var = np.zeros(self.rn2img.shape, dtype=float)

        #   - Subtract the overscan.  Uncertainty from the overscan subtraction
        #     is added to the variance.
        if self.par['use_overscan']:
            self.subtract_overscan()

        # TODO: Do we need to keep trim and orient as optional?

        #   - Trim to the data region.  This trims the image, rn2img, proc_var,
        #     var, and datasec_img.
        if self.par['trim']:
            self.trim()

        #   - Re-orient to PypeIt convention. This re-orients the image, rn2img,
        #     proc_var, var, and datasec_img.
        if self.par['orient']:
            self.orient()

        #   - Check the shape of the bpm
        if self.bpm.shape != self.image.shape:
            # The BPM is the wrong shape.  Assume this is because the
            # calibrations were taken with a different binning than the science
            # data, and assume this is because a BPM was provided as an
            # argument.  First erase the "protected" attribute, then access it
            # again using the @property method, which will recreated it based on
            # the expected shape for this frame.
            bpm_shape = self.bpm.shape                  # Save the current shape for the warning
            self._bpm = None                            # This erases the current bpm attribute
            if self.bpm.shape != self.image.shape:      # This recreates it
                # This should only happen because of a coding error, not a user error
                msgs.error(f'CODING ERROR: From-scratch BPM has incorrect shape!')
            # If the above was successful, the code can continue, but first warn
            # the user that the code ignored the provided bpm.
            msgs.warn(f'Bad-pixel mask has incorrect shape: found {bpm_shape}, expected '
                      f'{self.image.shape}.  Assuming this is because different binning used for '
                      'various frames.  Recreating BPM specifically for this frame '
                      f'({os.path.basename(self.filename)}) and assuming the difference in the '
                      'binning will be handled later in the code.')
            
        #   - Subtract master bias
        if self.par['use_biasimage']:
            # Bias frame.  Shape and orientation must match *processed* image,.
            # Uncertainty from the bias subtraction is added to the variance.
            self.subtract_bias(bias)

        #   - Subtract dark current.  The tabulated dark current is *always*
        #     subtracted regardless of whether or not use_darkimage is true.  If
        #     a dark frame is provided and subtracted, its shape and orientation
        #     must match *processed* image, and the units *must* be in
        #     electrons/counts.  If available, uncertainty from the dark
        #     subtraction is added to the variance.
        self.subtract_dark(dark_image=dark if self.par['use_darkimage'] else None)

        # TODO: We should be performing the cosmic-ray detection/masking here,
        # *before* flat-fielding, right?.

        # Calculate flexure, if slits are provided and the correction is
        # requested.  NOTE: This step must come after trim, orient (just like
        # bias and dark subtraction) and before field flattening.
        self.spat_flexure_shift = None if slits is None or not self.par['spat_flexure_correct'] \
                                    else self.spatial_flexure_shift(slits)

        # Flat-field the data.  This propagates the flat-fielding corrections to
        # the variance.  The returned bpm is propagated to the PypeItImage
        # bitmask below.
        flat_bpm = self.flatfield(flatimages, slits=slits, debug=debug) if self.use_flat else None

        # Calculate the inverse variance
        # TODO: This approach assumes that the dark-current provided/known for
        # each detector is more accurate and precise than using the counts in
        # the dark image itself (assuming one is available) to calculate the
        # dark-current shot noise.  We could pass ``darkcurr=dark`` if ``dark``
        # is available instead...
        self.ivar = self.build_ivar()

        # Generate a PypeItImage.
        # NOTE: To reconstruct the variance model, you need base_var, image,
        # img_scale, noise_floor, and shot_noise.
        pypeitImage = pypeitimage.PypeItImage(self.image, ivar=self.ivar, rn2img=self.rn2img,
                                              base_var=self.base_var, img_scale=self.img_scale,
                                              bpm=self.bpm, detector=self.detector,
                                              spat_flexure=self.spat_flexure_shift,
                                              PYP_SPEC=self.spectrograph.name,
                                              units='e-' if self.par['apply_gain'] else 'ADU',
                                              exptime=self.exptime,
                                              noise_floor=self.par['noise_floor'],
                                              shot_noise=self.par['shot_noise'])
        pypeitImage.rawheadlist = self.headarr
        pypeitImage.process_steps = [key for key in self.steps.keys() if self.steps[key]]

        # Mask(s)
        if self.par['mask_cr']:
            pypeitImage.build_crmask(self.par)
        nonlinear_counts = self.spectrograph.nonlinear_counts(self.detector,
                                                              apply_gain=self.par['apply_gain'])
        pypeitImage.build_mask(saturation=nonlinear_counts)
        if flat_bpm is not None:
            pypeitImage.update_mask('BADSCALE', indx=flat_bpm)

        # Return
        return pypeitImage

    def spatial_flexure_shift(self, slits, force=False):
        """
        Calculate a spatial shift in the edge traces due to flexure.

        This is a simple wrapper for
        :func:`~pypeit.core.flexure.spat_flexure_shift`.

        Args:
            slits (:class:`~pypeit.slittrace.SlitTraceSet`, optional):
                Slit edge traces
            force (:obj:`bool`, optional):
                Force the image to be field flattened, even if the step log
                (:attr:`steps`) indicates that it already has been.

        """
        step = inspect.stack()[0][3]
        if self.steps[step] and not force:
            # Already field flattened
            msgs.warn('Spatial flexure shift already calculated.')
            return
        self.spat_flexure_shift = flexure.spat_flexure_shift(self.image, slits)
        self.steps[step] = True

    def flatfield(self, flatimages, slits=None, force=False, debug=False):
        """
        Field flatten the processed image.

        This method uses the results of the flat-field modeling code (see
        :class:`~pypeit.flatfield.FlatField`) and any measured spatial shift due
        to flexure to construct slit-illumination, spectral response, and
        pixel-to-pixel response corrections, and multiplicatively removes them
        from the current image.  If available, the calculation is propagated to
        the variance image; however, no uncertainty in the flat-field
        corrections are included.

        .. warning::

            If you want the spatial flexure to be accounted for, you must first
            calculate the shift using
            :func:`~pypeit.images.rawimage.RawImage.spatial_flexure_shift`.

        Args:
            flatimages (:class:`~pypeit.flatfield.FlatImages`):
                Flat-field images used to apply flat-field corrections.
            slits (:class:`~pypeit.slittrace.SlitTraceSet`, optional):
                Used to construct the slit illumination profile, and only 
                required if this is to be calculated and normalized out.  See
                :func:`~pypeit.flatfield.FlatImages.fit2illumflat`.
            force (:obj:`bool`, optional):
                Force the image to be field flattened, even if the step log
                (:attr:`steps`) indicates that it already has been.
            debug (:obj:`bool`, optional):
                Run in debug mode.

        Returns:
            `numpy.ndarray`_: Returns a boolean array flagging pixels were the
            total applied flat-field value (i.e., the combination if the
            pixelflat and illumination corrections) was <=0.
        """
        step = inspect.stack()[0][3]
        if self.steps[step] and not force:
            # Already field flattened
            msgs.warn('Image was already flat fielded.')
            return

        # Check input
        if flatimages.pixelflat_norm is None:
            # We cannot do any flat-field correction without a pixel flat (yet)
            msgs.error("Flat fielding desired but not generated/provided.")
        if self.par['use_illumflat'] and slits is None:
            msgs.error('Need to provide slits to create illumination flat.')
        if self.par['use_specillum'] and flatimages.pixelflat_spec_illum is None:
            msgs.error("Spectral illumination correction desired but not generated/provided.")

        # Generate the illumination flat, as needed
        illum_flat = 1.0
        if self.par['use_illumflat']:
            illum_flat = flatimages.fit2illumflat(slits, flexure_shift=self.spat_flexure_shift)
            if debug:
                left, right = slits.select_edges(flexure=self.spat_flexure_shift)
                viewer, ch = display.show_image(illum_flat, chname='illum_flat')
                display.show_slits(viewer, ch, left, right)  # , slits.id)
                #
                orig_image = self.image.copy()
                viewer, ch = display.show_image(orig_image, chname='orig_image')
                display.show_slits(viewer, ch, left, right)  # , slits.id)

        # Apply the relative spectral illumination
        spec_illum = flatimages.pixelflat_spec_illum if self.par['use_specillum'] else 1.

        # Apply flat-field correction
        # NOTE: Using flat.flatfield to effectively multiply image*img_scale is
        # a bit overkill...
        total_flat = flatimages.pixelflat_norm * illum_flat * spec_illum
        self.img_scale = utils.inverse(total_flat)
        self.image, flat_bpm = flat.flatfield(self.image, total_flat)
        self.steps[step] = True
        return flat_bpm

    def orient(self, force=False):
        """
        Orient image attributes such that they follow the ``PypeIt`` convention
        with spectra running blue (down) to red (up) and with orders decreasing
        from high (left) to low (right).
        
        This edits :attr:`image`, :attr:`rn2img` (if it exists),
        :attr:`proc_var` (if it exists), and :attr:`datasec_img` in place.

        Args:
            force (:obj:`bool`, optional):
                Force the image to be re-oriented, even if the step log
                (:attr:`steps`) indicates that it already has been.
        """
        step = inspect.stack()[0][3]
        # Check if already oriented
        if self.steps[step] and not force:
            msgs.warn('Image was already oriented.')
            return
        # Orient the image to have blue/red run bottom to top
        self.image = self.spectrograph.orient_image(self.detector, self.image)
        if self.rn2img is not None:
            self.rn2img = self.spectrograph.orient_image(self.detector, self.rn2img)
        if self.proc_var is not None:
            self.proc_var = self.spectrograph.orient_image(self.detector, self.proc_var)
        self.datasec_img = self.spectrograph.orient_image(self.detector, self.datasec_img)
        self.steps[step] = True

    def subtract_bias(self, bias_image, force=False):
        """
        Subtract a bias image.
        
        If the ``bias_image`` object includes an inverse variance image and if
        :attr:`var` is available, the error in the bias is propagated to the
        bias-subtracted image.

        Args:
            bias_image (:class:`~pypeit.images.pypeitimage.PypeItImage`):
                Bias image
            force (:obj:`bool`, optional):
                Force the image to be subtracted, even if the step log
                (:attr:`steps`) indicates that it already has been.
        """
        step = inspect.stack()[0][3]
        if self.steps[step] and not force:
            # Already bias subtracted
            msgs.warn('Image was already bias subtracted.')
            return
        self.image -= bias_image.image
        # TODO: Also incorporate the mask?
        if bias_image.ivar is not None and self.proc_var is not None:
            self.proc_var += utils.inverse(bias_image.ivar)
        self.steps[step] = True

    # TODO: expscale is currently not a parameter that the user can control.
    # Should it be?
    def subtract_dark(self, dark_image=None, expscale=True, force=False):
        """
        Subtract detector dark current.

        This method subtracts both the tabulated dark-current for the detector
        and a dark image.  For this to be appropriate, the dark image (if
        provided) *must* also have had the tabulated dark-current value
        subtracted from it.

        Also, the processing of the dark image (if provided) should match the
        processing of the image being processed.  For example, if this image has
        been bias subtracted, so should be the dark image.

        If the ``dark_image`` object includes an inverse variance estimate and
        if :attr:`proc_var` is available, the uncertainty in the dark will be
        propagated to the dark-subtracted image.

        .. warning::

            The current default automatically scales the dark frame to match the
            exposure time of the image being processed.  Typically dark frames
            should have the same exposure time as the image being processed, so
            this will have no effect.  However, beware if that's not the case,
            and make sure this scaling is appropriate.  Use ``expscale`` to
            turn it off.

        Args:
            dark_image (:class:`~pypeit.images.pypeitimage.PypeItImage`, optional):
                The dark image to subtract.  If None, only the tabulated
                dark-current is subtracted from *all* pixels.
            expscale (:obj:`bool`, optional):
                Scale the dark image (if provided) by the ratio of the exposure
                times so that the counts per second represented by the dark
                image are appropriately subtracted from image being processed.
            force (:obj:`bool`, optional):
                Force the dark to be subtracted, even if the step log
                (:attr:`steps`) indicates that it already has been.
        """
        step = inspect.stack()[0][3]
        if self.steps[step] and not force:
            # Already bias subtracted
            msgs.warn('Image was already dark subtracted.')
            return
        # TODO: Is the dark-current amplifier dependent?  Also, this usage means
        # that darkcurr cannot be None.
        # Tabulated dark current is in e-/pixel/hour and exptime is in s, the
        # 3600 factor converts the dark current to e-/pixel/s.
        npix = np.prod(parse.parse_binning(self.detector.binning))
        dark_count = npix*self.detector['darkcurr'] * self.exptime / 3600.
        if dark_image is not None:
            # TODO: Include a warning when the scaling is "signficant"?
            scale = self.exptime / dark_image.exptime if expscale else 1.
            # Warn the user if the counts in the dark image are significantly
            # different from the tabulated value.  The 50% difference is not
            # well justified but, for now, hard-coded.
            med_dark = np.median(scale * dark_image.image)
            if np.absolute(med_dark) > 0.5*dark_count:
                msgs.warn(f'Dark-subtracted dark frame has significant signal remaining.  Median '
                          f'count is {med_dark:.2f}; warning threshold = +/-{0.5*dark_count:.2f}.')
            # NOTE: This converts dark_count from a scalar to an array
            dark_count += scale * dark_image.image
            if dark_image.ivar is not None and self.proc_var is not None:
                # Include the variance in the dark image
                # TODO: Also incorporate the mask?
                self.proc_var += scale**2 * utils.inverse(dark_image.ivar)
        self.image -= dark_count
        self.steps[step] = True

    def subtract_overscan(self, force=False):
        """
        Analyze and subtract the overscan from the image

        Args:
            force (:obj:`bool`, optional):
                Force the image to be overscan subtracted, even if the step log
                (:attr:`steps`) indicates that it already has been.
        """
        step = inspect.stack()[0][3]
        if self.steps[step] and not force:
            # Already overscan subtracted
            msgs.warn("Image was already overscan subtracted!")
            return
        # Subtract the overscan.  var is the variance in the overscan
        # subtraction
        self.image, var = procimg.subtract_overscan(self.image, self.datasec_img,
                                                    self.oscansec_img,
                                                    method=self.par['overscan_method'],
                                                    params=self.par['overscan_par'],
                                                    var=self.rn2img)
        # Parse the returned value
        if self.rn2img is not None:
            # Include the variance in the overscan subtraction in the
            # "processing variance"
            self.proc_var += var
        self.steps[step] = True

    def subtract_pattern(self):
        """
        Analyze and subtract the pattern noise from the image.

        This is primarily a wrapper for
        :func:`~pypeit.core.procimg.subtract_pattern`.

        """
        step = inspect.stack()[0][3]
        if self.steps[step]:
            # Already pattern subtracted
            msgs.warn("Image was already pattern subtracted!")
            return

        # The image must have an overscan region for this to work.
        if not np.any(self.oscansec_img > 0):
            msgs.error('Image has no overscan region.  Pattern noise cannot be subtracted.')

        # The image cannot have already been trimmed
        if self.oscansec_img.shape != self.image.shape:
            msgs.error('Must estimate readnoise before trimming the image.')

        # Grab the frequency, if it exists in the header.  For some instruments,
        # PYPEITFRQ is added to the header in get_rawimage() in the spectrograph
        # file.  See keck_kcwi.py for an example.
        frequency = []
        try:
            # Grab a list of all the amplifiers
            amps = np.sort(np.unique(self.oscansec_img[np.where(self.oscansec_img > 0)]))
            for amp in amps:
                frequency.append(self.hdu[0].header['PYPFRQ{0:02d}'.format(amp)])
            # Final check to make sure the list isn't empty (which it shouldn't be, anyway)
            if len(frequency) == 0:
                frequency = None
        except KeyError:
            frequency = None
        # Subtract the pattern and overwrite the current image
        self.image = procimg.subtract_pattern(self.image, self.datasec_img, self.oscansec_img,
                                              frequency=frequency)
        self.steps[step] = True

    def trim(self, force=False):
        """
        Trim image attributes to include only the science data.
        
        This edits :attr:`image`, :attr:`rn2img` (if it exists),
        :attr:`proc_var` (if it exists), and :attr:`datasec_img` in place.

        Args:
            force (:obj:`bool`, optional):
                Force the image to be trimmed, even if the step log
                (:attr:`steps`) indicates that it already has been.
        """
        step = inspect.stack()[0][3]
        if self.rawimage.shape is not None and self.image.shape != self.rawimage.shape:
            # Image *must* have been trimmed already because shape does not
            # match raw image
            self.steps[step] = True
            msgs.warn('Image shape does not match original.')
            return
        if self.steps[step] and not force:
            # Already trimmed
            msgs.warn('Image was already trimmed.')
            return
        self.image = procimg.trim_frame(self.image, self.datasec_img < 1)
        if self.rn2img is not None:
            self.rn2img = procimg.trim_frame(self.rn2img, self.datasec_img < 1)
        if self.proc_var is not None:
            self.proc_var = procimg.trim_frame(self.proc_var, self.datasec_img < 1)
        # NOTE: This must be done last because the untrimmed image is used for
        # deciding what to trim!
        self.datasec_img = procimg.trim_frame(self.datasec_img, self.datasec_img < 1)
        self.steps[step] = True

    def __repr__(self):
        return ('<{:s}: file={}, steps={}>'.format(self.__class__.__name__, self.filename,
                                                   self.steps))


